from ._tools import *

def testWhiteningNode():
    vars = 5
    dim = (10000,vars)
    mat,mix,inp = get_random_mix(mat_dim=dim, avg=uniform(vars))
    w = mdp.nodes.WhiteningNode()
    w.train(inp)
    out = w.execute(inp)
    assert_array_almost_equal(mean(out, axis=0),
                              numx.zeros(dim[1]), decimal)
    assert_array_almost_equal(std(out, axis=0),
                              numx.ones(dim[1]), decimal - 3)

def testWhiteningNode_SVD():
    vars = 5
    dim = (10000,vars)
    mat,mix,inp = get_random_mix(mat_dim=dim, avg=uniform(vars))
    w = mdp.nodes.WhiteningNode(svd=True)
    w.train(inp)
    out = w.execute(inp)

    assert_array_almost_equal(mean(out, axis=0),
                              numx.zeros(dim[1]), decimal)
    assert_array_almost_equal(std(out, axis=0),
                              numx.ones(dim[1]), decimal - 3)




"""
Module for HTML slideshows.

It uses the templating library 'Templet'.

The slideshow base class HTMLSlideShow does not display anything, but can
be used to derive custom slideshows like in BiMDP.

The JavaScript slideshow code in this module was originally inspired by a
slideshow script found at
http://javascript.internet.com/miscellaneous/image-slideshow.html
(which in turn seems to be based on something from http://www.ricocheting.com)
"""

from builtins import str

import random
import tempfile
import os
import webbrowser
import warnings

from . import templet

_BASIC_CSS_FILENAME = "basic.css"
_SLIDESHOW_CSS_FILENAME = "slideshow.css"

def basic_css():
    """Return the basic default CSS."""
    css_filename = os.path.join(os.path.split(__file__)[0],
                                _BASIC_CSS_FILENAME)
    with open(css_filename, 'r') as css_file:
        css = css_file.read()
    return css

def slideshow_css():
    """Return the additional CSS for a slideshow."""
    css_filename = os.path.join(os.path.split(__file__)[0],
                                _SLIDESHOW_CSS_FILENAME)
    with open(css_filename, 'r') as css_file:
        css = css_file.read()
    return css


class HTMLSlideShow(templet.Template):
    """Abstract slideshow base class.

    It does not display anything, but can be adapted by overriding
    some of the templating attributes. See ImageHTMLSlideShow for an example.
    """

    def __init__(self, title=None, delay=100, delay_delta=20,
                 loop=True, slideshow_id=None, shortcuts=True, **kwargs):
        """Return the complete HTML code for the slideshow.

        title -- Optional slideshow title (for defualt None not title is shown).
        delay - Delay between slides in ms (default 100).
        delay_delta - Step size for increasing or decreasing the delay.
        loop -- If True continue with first slide when the last slide is
            reached during the automatic slideshow (default is False).
        slideshow_id -- String with the id used for the JS closure, and this
            is also the id of the div with the slideshow (so it can be used
            by CSS) and it is used as a prefix for the HTML elements.
            If the value is None (default) then a random id is used.
        shortcuts -- Bind keyboard shortcuts to this slideshow (default is
            True). Note that keyboard shortcuts only work for a single
            slideshow per page.
        """
        # translate boolean variable into JS format
        if loop:
            loop = "true"
        else:
            loop = "false"
        if slideshow_id is None:
            slideshow_id = self._get_random_id()
        self.slideshow_id = slideshow_id
        kwargs.update(vars())
        del kwargs["self"]
        super(HTMLSlideShow, self).__init__(**kwargs)

    def _get_random_id(self):
        """Factory method for random slideshow id."""
        return "slideshow%d" % random.randint(10000, 99999)


    template = r'''
<script language="JavaScript">
<!-- Begin

// use closure to create object, the slideshow_id is the only global variable
var $slideshow_id = function () {

    var that = {};

    var current_slide = 0; // current slide index
    var show_delay = $delay; // delay in milliseconds
    var loop_slideshow = $loop; // loop in auto mode
    // shortcuts to form elements, initialized in onLoad
    var slideform;
    var slideselect;

    $<js_controls_template>
    $<js_loadslide_template>
    $<js_update_template>
    $<js_onload_template>

    that.onSelectorChange = function () {
        current_slide = slideselect.selectedIndex;
        that.updateSlide();
    }

    that.next = function () {
        if (slideselect[current_slide+1]) {
            current_slide += 1;
            that.updateSlide();
        }
        else if (loop_slideshow) {
            that.first();
        }
    }

    that.previous = function () {
        if (current_slide-1 >= 0) {
            current_slide -= 1;
            that.updateSlide();
        }
        else if (loop_slideshow) {
            that.last();
        }
    }

    that.first = function () {
        current_slide = 0;
        that.updateSlide();
    }

    that.last = function () {
        current_slide = slideselect.length-1;
        that.updateSlide();
    }

    // start or stop the slideshow
    that.startstop = function (text) {
        if (text === "Start") {
            slideform.startbutton.value = "Stop";
            if (!loop_slideshow &&
                (current_slide === slideselect.length-1)) {
                // restart slideshow
                current_slide = -1;  // is directly increased by showAuto
            }
            that.showAuto();
        } else {
            slideform.startbutton.value = "Start";
        }
    }

    // continuously show the slideshow
    that.showAuto = function () {
        if (slideform.startbutton.value == "Stop") {
            if (current_slide == slideselect.length-1) {
                if (loop_slideshow) {
                    current_slide = 0;
                    that.updateSlide();
                } else {
                    slideform.startbutton.value = "Start";
                }
            } else {
                current_slide = current_slide+1;
                that.updateSlide();
            }
            window.setTimeout("$slideshow_id.showAuto()", show_delay);
        }
    }

    // end of closure, return created object
    return that;
}();

${{
if shortcuts:
    self.js_keyboard_shortcuts_template(vars())
}}

//  End -->
</script>

<div class="slideshow" id="$slideshow_id">

$<html_top_template>

<form name=${slideshow_id}_slideform>
<table class="slideshow">
${{
if title:
    self.write('<tr><td><b> %s </b></td></tr>' % title)
}}

$<html_box_template>

<tr><td>
<select name="${slideshow_id}_slideselect"
    onChange="$slideshow_id.onSelectorChange();">
<option value="${filenames[0]}" selected>${filenames[0]}
${{
for filename in filenames[1:]:
    self.write('<option value="%s">%s\n' % (filename, filename))
}}
</select>
</td></tr>
<tr><td>

$<html_buttons_template>

</td></tr>

$<html_controls_template>

</table>
</form>

$<html_bottom_template>

</div>

<SCRIPT LANGUAGE="JavaScript">
<!-- Begin
$slideshow_id.onLoad();
//  End -->
</script>
'''

    js_controls_template = r'''
    // step size for in- or decreasing the delay
    var delay_delta = $delay_delta;

    that.slower = function () {
        show_delay += delay_delta;
        slideform.${slideshow_id}_delaytext.value = show_delay.toString();
    }

    that.faster = function (text) {
        show_delay -= delay_delta;
        if (show_delay < 0) {
            show_delay = 0;
        }
        slideform.${slideshow_id}_delaytext.value = show_delay.toString();
    }

    that.changeDelay = function () {
        var new_delay = parseInt(slideform.${slideshow_id}_delaytext.value, 10);
        if (new_delay < 0) {
             new_delay = 0;
            }
        show_delay = new_delay;
        slideform.${slideshow_id}_delaytext.value = new_delay.toString();
    }
'''

    js_update_template = r'''
    that.updateSlide = function () {
        slideselect.selectedIndex = current_slide;
        that.loadSlide();
    }
'''

    # overwrite this to implement the actual slide change
    js_loadslide_template = r'''
    that.loadSlide = function () {
    }
'''

    js_onload_template = r'''
    that.onLoad = function () {
        slideform = document.${slideshow_id}_slideform;
        slideselect = slideform.${slideshow_id}_slideselect;
        current_slide = slideselect.selectedIndex;
        that.updateSlide();
        slideform.${slideshow_id}_delaytext.value = show_delay.toString();
    }
'''

    # define keyboard shortcuts,
    # note that these are also mentionend in the button hover-text
    js_keyboard_shortcuts_template = r'''
document.onkeydown = function(e) {
    if (!e.ctrlKey) {  // control key must be pressed
        return;
    }
    else if (e.which == 37) { // left key
        document.getElementById("${slideshow_id}_prevButton").click();
    }
    else if(e.which == 39) { // right key
        document.getElementById("${slideshow_id}_nextButton").click();
    }
    else if(e.which == 38) { // up key
        document.getElementById("${slideshow_id}_firstButton").click();
    }
    else if(e.which == 40) { // down key
        document.getElementById("${slideshow_id}_lastButton").click();
    }
    else if(e.which == 45) { // insert key
        document.getElementById("${slideshow_id}_startButton").click();
    }
}
'''

    html_buttons_template = r'''
<input type=button onClick="$slideshow_id.first();"
    value="|<<" title="First" id="${slideshow_id}_firstButton">
<input type=button onClick="$slideshow_id.previous();"
    value="<" title="Previous (Ctrl+Left)" id="${slideshow_id}_prevButton">
<input type=button name="startbutton"
    onClick="$slideshow_id.startstop(this.value);"
    value="Start" title="Autoplay (Ctrl+Insert)"
    id="${slideshow_id}_startButton">
<input type=button onClick="$slideshow_id.next();" value=">"
    title="Next (Ctrl+Right)"
    id="${slideshow_id}_nextButton">
<input type=button onClick="$slideshow_id.last();" value=">>|" title="Last"
    id="${slideshow_id}_lastButton">
'''

    html_controls_template = r'''
${{
if delay is not None:
    self.write('<tr><td>\n')
    self.html_delay_template(vars())
    self.write('</td></tr>\n')
}}
'''

    html_delay_template = r'''
delay: <input type="text" name="${slideshow_id}_delaytext"
    onChange="$slideshow_id.changeDelay();" value="0" size="4"> ms
<input type=button onClick="$slideshow_id.faster();" value="-" title="Faster">
<input type=button onClick="$slideshow_id.slower();" value="+" title="Slower">
'''

    html_top_template = r'''
'''

    html_box_template = r'''
'''

    html_bottom_template = r'''
'''


class SectionHTMLSlideShow(HTMLSlideShow):
    """Astract slideshow with additional support for section markers."""

    def __init__(self, section_ids, slideshow_id=None, **kwargs):
        """Return the complete HTML code for the slideshow.

        section_ids -- List with the section id for each slide index. The id
            can be a string or a number.

        For additional keyword arguments see the super class.
        """
        # we need the slideshow_id for the section names
        if slideshow_id is None:
            slideshow_id = self._get_random_id()
        kwargs.update(vars())
        # check if there is more than one section slideshow_id,
        # otherwise some controls must be disabled to prevent infinite loop
        only_one_section = "false"
        first_section_id = section_ids[0]
        for section_id in section_ids:
            if section_id != first_section_id:
                break
        else:
            only_one_section = "true"
        kwargs["only_one_section"] = only_one_section
        # translate section_id_list into JavaScript list
        section_ids = [str(section_id) for section_id in section_ids]
        js_section_ids = "".join(['        "%s_section_id_%s",\n' %
                                  (slideshow_id, section_id)
                                  for section_id in section_ids])
        js_section_ids = "\n" + js_section_ids[:-2]
        kwargs["js_section_ids"] = js_section_ids
        del kwargs["self"]
        super(SectionHTMLSlideShow, self).__init__(**kwargs)


    js_update_template = r'''
    // maps slide index to section slideshow_id
    var section_ids = new Array($js_section_ids);
    // currently highlighted section slideshow_id
    var current_section_id = section_ids[0];

    that.updateSlide = function () {
        document.getElementById(current_section_id).className =
            "inactive_section";
        current_section_id = section_ids[current_slide]
        document.getElementById(current_section_id).className =
            "active_section";
        slideselect.selectedIndex = current_slide;
        that.loadSlide();
    }

    // use this function when a section is selected,
    // e.g. onClick="setSlide(42)"
    that.setSlide = function (index) {
        current_slide = index;
        that.updateSlide();
    }

    that.previousSection = function () {
        if ($only_one_section) {
            return;
        }
        while (current_section_id === section_ids[current_slide]) {
            if (current_slide > 0) {
                current_slide -= 1;
            } else {
                current_slide = slideselect.length-1;
            }
        }
        var new_section_id = section_ids[current_slide];
        // now go to start of this section
        while (new_section_id === section_ids[current_slide]) {
            current_slide -= 1;
            if (current_slide < 0) {
                break;
            }
        }
        current_slide += 1;
        that.updateSlide();
    }

    that.nextSection = function () {
        if ($only_one_section) {
            return;
        }
        while (current_section_id === section_ids[current_slide]) {
            if (current_slide+1 < slideselect.length) {
                current_slide += 1;
            } else {
                current_slide = 0;
            }
        }
        that.updateSlide();
    }

    $<js_loadslide_template>
'''

    # define keyboard shortcuts,
    # note that these are also mentionend in the button hover-text
    js_keyboard_shortcuts_template = r'''
document.onkeydown = function(e) {
    if (!e.ctrlKey) { // control key must be pressed
        return;
    }
    else if (e.which === 37) { // left key
        document.getElementById("${slideshow_id}_prevButton").click();
    }
    else if(e.which === 39) { // right key
        document.getElementById("${slideshow_id}_nextButton").click();
    }
    else if(e.which === 38) { // up key
        document.getElementById("${slideshow_id}_prevSectionButton").click();
    }
    else if(e.which === 40) { // down key
        document.getElementById("${slideshow_id}_nextSectionButton").click();
    }
    else if(e.which === 45) { // insert key
        document.getElementById("${slideshow_id}_startButton").click();
    }
}
    '''

    html_buttons_template = r'''
<input type=button onClick="$slideshow_id.first();"
    value="|<<" title="First" id="${slideshow_id}_firstButton">
<input type=button onClick="$slideshow_id.previousSection();" value="|<"
    title="Previous Section (Ctrl+Up)" id="${slideshow_id}_prevSectionButton">
<input type=button onClick="$slideshow_id.previous();"
    value="<" title="Previous (Ctrl+Left)" id="${slideshow_id}_prevButton">
<input type=button name="startbutton"
    onClick="$slideshow_id.startstop(this.value);"
    value="Start" title="Autoplay (Ctrl+Insert)"
    id="${slideshow_id}_startButton">
<input type=button onClick="$slideshow_id.next();" value=">"
    title="Next (Ctrl+Right)"
    id="${slideshow_id}_nextButton">
<input type=button onClick="$slideshow_id.nextSection();" value=">|"
    title="Next Section (Ctrl+Down)" id="${slideshow_id}_nextSectionButton">
<input type=button onClick="$slideshow_id.last();" value=">>|" title="Last"
    id="${slideshow_id}_lastButton">
'''

    html_controls_template = r'''
${{super(SectionHTMLSlideShow, self).html_controls_template(vars())}}

<tr><td>
<div id="${slideshow_id}_sections_panel"
    style="margin-left:auto; margin-right: auto;">
${{
last_section_id = None
link = ''
for index, section_id in enumerate(section_ids):
    if section_id != last_section_id:
        if index > 0:
            self.write(link + '&nbsp;| ')
        last_section_id = section_id
        link = ('<span class="inactive_section" ' +
                'id="%s_section_id_%s" ' % (slideshow_id, section_id) +
                'onClick="%s.setSlide(%d);">%s</span>' %
                    (slideshow_id, index, section_id))
self.write(link + '\n')
}}
</div>
</td></tr>
'''


def image_slideshow_css():
    """Use nearest neighbour resampling in Firefox 3.6+ and IE.

    Webkit (Chrome, Safari) does not support this yet.
    (see http://code.google.com/p/chromium/issues/detail?id=1502)
    """
    return slideshow_css() + '''
img.slideshow {
    image-rendering: -moz-crisp-edges;
    -ms-interpolation-mode: nearest-neighbor;
}
'''


class ImageHTMLSlideShow(HTMLSlideShow):
    """Slideshow for images.

    This also serves as an example for implementing a slideshow based on
    HTMLSlideShow.
    """

    def __init__(self, filenames, image_size,
                 magnification=1, mag_control=True, **kwargs):
        """Return the complete HTML code for a slideshow of the given images.

        filenames -- sequence of strings, containing the path for each image
        image_size -- Tuple (x,y) with the original image size, or enter
            a different size to force scaling.
        magnification -- Magnification factor for images (default 1). This
            factor is applied on top of the provided image size.
        mag_control -- Set to True (default) to display a magnification control
            element.

        For additional keyword arguments see the super class.
        """
        if len(filenames) == 0:
            raise Exception("Empty list was given.")
        kwargs.update(vars())
        # translate image size to width and heigh to be used in the templates
        del kwargs["image_size"]
        kwargs["width"] = image_size[0]
        kwargs["height"] = image_size[1]
        del kwargs["self"]
        super(ImageHTMLSlideShow, self).__init__(**kwargs)

    js_controls_template = r'''
    ${{super(ImageHTMLSlideShow, self).js_controls_template(vars())}}

    var magnification = $magnification; // image magnification
    var original_width = $width; // original image width
    var original_height = $height; // original image height

    that.smaller = function () {
        magnification = magnification / 2;
        slideform.${slideshow_id}_magtext.value = magnification.toString();
        that.resizeImage();
    }

    that.larger = function (text) {
        magnification = magnification * 2;
        slideform.${slideshow_id}_magtext.value = magnification.toString();
        that.resizeImage();
    }

    that.changeMag = function () {
        magnification = parseFloat(slideform.${slideshow_id}_magtext.value);
        that.resizeImage();
    }

    $<js_controls_resize_template>
'''

    js_controls_resize_template = r'''
    that.resizeImage = function () {
        document.images.${slideshow_id}_image_display.width =
            parseInt(magnification * original_width, 10);
        document.images.${slideshow_id}_image_display.height =
            parseInt(magnification * original_height, 10);
    }
'''

    js_loadslide_template = r'''
    that.loadSlide = function () {
        document.images.${slideshow_id}_image_display.src =
            slideselect[current_slide].value;
    }
'''

    js_onload_template = r'''
    that.onLoad = function () {
        slideform = document.${slideshow_id}_slideform;
        slideselect = slideform.${slideshow_id}_slideselect;
        current_slide = slideselect.selectedIndex;
        that.updateSlide();
        ${{
        if delay is not None:
            self.write('slideform.%s_delaytext.value = ' % slideshow_id +
                       'show_delay.toString();\n')
        }}
        ${{
        if mag_control:
            self.write('slideform.%s_magtext.value = ' % slideshow_id +
                       'magnification.toString();\n')
        }}
        that.resizeImage();
    }
'''

    html_box_template = r'''
<tr>
<td style="padding: 20 20 20 20">
<img class="slideshow" src="" name="${slideshow_id}_image_display"
    width="$width" height="$height">
</td>
</tr>
'''

    html_controls_template = r'''
${{
if mag_control or (delay is not None):
    self.write('<tr><td align=center>\n')
    if mag_control:
        self.html_mag_template(vars())
        if delay is not None:
            self.write('<br>\n')
    if delay is not None:
        self.html_delay_template(vars())
    self.write('</td></tr>\n')
}}
'''

    html_mag_template = r'''
magnification: <input type="text" name="${slideshow_id}_magtext"
    onChange="$slideshow_id.changeMag();" value="0" size="2">
<input type=button onClick="$slideshow_id.smaller();" value="-" title="Smaller">
<input type=button onClick="$slideshow_id.larger();" value="+" title="Larger">
'''


class SectionImageHTMLSlideShow(SectionHTMLSlideShow, ImageHTMLSlideShow):
    """Image slideshow with section markers."""

    def __init__(self, filenames, section_ids, image_size, **kwargs):
        """Return the HTML code for a sectioned slideshow of the given images.

        For keyword arguments see the super classes.
        """
        if len(section_ids) != len(filenames):
            err = ("The number of section slideshow_id entries does not match "
                   "the number of slides / filenames.")
            raise Exception(err)
        kwargs.update(vars())
        del kwargs["self"]
        super(SectionImageHTMLSlideShow, self).__init__(**kwargs)

    js_controls_resize_template = r'''
    that.resizeImage = function () {
        document.images.${slideshow_id}_image_display.width =
            parseInt(magnification * original_width, 10);
        document.images.${slideshow_id}_image_display.height =
            parseInt(magnification * original_height, 10);
        // make sure that section ids are nicely line wrapped
        var section_panel_width = 250;
        if (magnification * original_height > section_panel_width) {
            section_panel_width = magnification * original_width;
        }
        document.getElementById("${slideshow_id}_sections_panel").style.width =
            parseInt(section_panel_width, 10) + "px";
    }
'''


### helper functions ###

# TODO: extract image size automatically,
#    but this introduces an optional dependency on PIL

def image_slideshow(filenames, image_size, title=None, section_ids=None,
                    delay=100, delay_delta=20, loop=True, slideshow_id=None,
                    magnification=1, mag_control=True, shortcuts=True):
    """Return a string with the JS and HTML code for an image slideshow.

    Note that the CSS code for the slideshow is not included, so you should
    add SLIDESHOW_STYLE or a custom style to your CSS code.

    filenames -- Sequence of the image filenames.
    image_size -- Tuple (x,y) with the original image size, or enter
        a different size to force scaling.
    title -- Optional slideshow title (for default None not title is shown).
    section_ids -- List with the section id for each slide index. The id
            can be a string or a number. Default value None disables the
            section feature.

    For additional keyword arguments see the ImageHTMLSlideShow class.
    """
    if section_ids:
        slideshow = SectionImageHTMLSlideShow(**vars())
    else:
        slideshow = ImageHTMLSlideShow(**vars())
    return str(slideshow)

def show_image_slideshow(filenames, image_size, filename=None, title=None,
                         section_ids=None, delay=100, delay_delta=20,
                         loop=True, slideshow_id=None,
                         magnification=1, mag_control=True, open_browser=True):
    """Write the slideshow into a HTML file, open it in the browser and
    return a file object pointing to the file. If the filename is not given,
    a temporary file is used, and will be deleted when the returned file object
    is closed or destroyed.

    filenames -- Sequence of the image filenames.
    image_size -- Tuple (x,y) with the original image size, or enter
        a different size to force scaling.
    filename -- Filename for the HTML file to be created. If None
            a temporary file is created.
    title -- Optional slideshow title (for default None not title is shown).
    section_ids -- List with the section id for each slide index. The id
            can be a string or a number. Default value None disables the
            section feature.
    open_browser -- If True (default value) then the slideshow file is
        automatically opened in a webbrowser. One can also use string value
        with the browser name (for webbrowser.get) to request a specific
        browser.

    For additional keyword arguments see the ImageHTMLSlideShow class.
    """
    if filename is None:
        html_file = tempfile.NamedTemporaryFile(suffix=".html", prefix="MDP_")
    else:
        html_file = open(filename, 'w')
    html_file.write('<html>\n<head>\n<title>%s</title>\n' % title)
    html_file.write('<style type="text/css" media="screen">')
    html_file.write(basic_css() + image_slideshow_css())
    html_file.write('</style>\n</head>\n<body>\n')
    kwargs = vars()
    del kwargs['filename']
    del kwargs['open_browser']
    del kwargs['html_file']
    html_file.write(image_slideshow(**kwargs))
    html_file.write('</body>\n</html>')
    html_file.flush()

    if open_browser:
        if isinstance(open_browser, str):
            try:
                custom_browser = webbrowser.get(open_browser)
                custom_browser.open(os.path.abspath(filename))
            except webbrowser.Error:
                err = ("Could not open browser '%s', using default." %
                       open_browser)
                warnings.warn(err)
                webbrowser.open(os.path.abspath(filename))
        else:
            webbrowser.open(os.path.abspath(filename))
    return html_file



from builtins import range
from ._tools import *
import py.test

requires_signal = skip_on_condition(
    "not hasattr(mdp.nodes, 'Convolution2DNode')",
    "This test requires the 'scipy.signal' module.")

@requires_signal
def testConvolution2Dsimple():
    # copied over from convolution_nodes.py
    im = numx_rand.rand(4, 3,3)
    node = mdp.nodes.Convolution2DNode(numx.array([[[1.]]]))
    node.execute(im)

@requires_signal
def testConvolution2DNodeFunctionality():
    filters = numx.empty((3,1,1))
    filters[:,0,0] = [1.,2.,3.]
    x = numx_rand.random((10,3,4))
    
    for mode in ['valid', 'same', 'full']:
        for boundary in ['fill', 'wrap', 'symm']:
            node = mdp.nodes.Convolution2DNode(filters, approach='linear', mode=mode,
                                               boundary=boundary, output_2d=False)
            y = node.execute(x)
            
            assert_equal(y.shape, (x.shape[0], 3, x.shape[1], x.shape[2]))
            for n_flt in range(3):
                assert_array_equal(x*(n_flt+1.), y[:,n_flt,:,:])

@requires_signal
def testConvolution2DNode_2D3Dinput():
    filters = numx.empty((3,1,1))
    filters[:,0,0] = [1.,2.,3.]
    
    # 1) input 2D/3D
    x = numx_rand.random((10,12))        
    node = mdp.nodes.Convolution2DNode(filters, approach='linear',
                                       input_shape=(3,4), output_2d=False)
    y = node.execute(x)
    assert_equal(y.shape, (x.shape[0], 3, 3, 4))
    
    x = numx.random.random((10,3,4)) 
    node = mdp.nodes.Convolution2DNode(filters, output_2d=False)
    y = node.execute(x)
    assert_equal(y.shape, (x.shape[0], 3, 3, 4))
        
    # 2) output 2D/3D
    x = numx.random.random((10,12))        
    node = mdp.nodes.Convolution2DNode(filters, approach='linear',
                                       input_shape=(3,4), output_2d=True)
    y = node.execute(x)
    assert_equal(y.shape, (x.shape[0], 3*3*4))
    for i in range(3):
        assert_array_equal(x*(i+1.), y[:,i*12:(i+1)*12])

@requires_signal
def testConvolution2DNode_fft():
    filters = numx.empty((3,1,1))
    filters[:,0,0] = [1.,2.,3.]
    x = numx.random.random((10,3,4))
    
    for mode in ['valid', 'same', 'full']:
        node = mdp.nodes.Convolution2DNode(filters, approach='fft', mode=mode,
                                           output_2d=False)
        y = node.execute(x)

        assert_equal(y.shape, (x.shape[0], 3, x.shape[1], x.shape[2]))
        for n_flt in range(3):
            assert_array_almost_equal(x*(n_flt+1.), y[:,n_flt,:,:], 6)

    # with random filters
    x = numx.random.random((10,30,20))
    filters = numx.random.random((3,5,4))
    for mode in ['valid', 'same', 'full']:
        node_fft = mdp.nodes.Convolution2DNode(filters, approach='fft', mode=mode,
                                               output_2d=False)
        node_lin = mdp.nodes.Convolution2DNode(filters, approach='linear', mode=mode,
                                               boundary='fill', output_2d=False)
        y_fft = node_fft.execute(x)
        y_lin = node_lin.execute(x)
        
        assert_array_almost_equal(y_fft, y_lin, 6)

@requires_signal
def testConvolution2DNode_in_Flow():
    filters = numx.empty((3,1,1))
    filters[:,0,0] = [1.,2.,3.]
    
    # with 3D input
    x = numx.random.random((10,3,4))   
    node = mdp.nodes.Convolution2DNode(filters, output_2d=True)
    flow = mdp.Flow([node, mdp.nodes.PCANode(output_dim=3)])
    flow.train(x)
    flow.execute(x)
    
    # with 2D input
    x = numx.random.random((10,12))
    node = mdp.nodes.Convolution2DNode(filters, input_shape=(3,4), output_2d=True)
    flow = mdp.Flow([mdp.nodes.IdentityNode(),
                     node,
                     mdp.nodes.PCANode(output_dim=3)])
    flow.train(x)
    flow.execute(x)

@requires_signal
def testConvolution2DNode_arguments():
    # filters must be 3D
    filters = numx.random.random((5,4))
    py.test.raises(mdp.NodeException,
                   "mdp.nodes.Convolution2DNode(filters)")
    filters = numx.random.random((5,4,2,2))
    py.test.raises(mdp.NodeException,
                   "mdp.nodes.Convolution2DNode(filters)")
    # filters must be array
    filters = [[[2.]]]
    py.test.raises(mdp.NodeException,
                   "mdp.nodes.Convolution2DNode(filters)")
 
    filters = numx.random.random((1,1,1))
    with py.test.raises(mdp.NodeException):
        mdp.nodes.Convolution2DNode(filters, approach='bug')
    with py.test.raises(mdp.NodeException):
        mdp.nodes.Convolution2DNode(filters, mode='bug')
    with py.test.raises(mdp.NodeException):
        mdp.nodes.Convolution2DNode(filters, boundary='bug')

@requires_signal
def testConvolution2DNode_shape_mismatch():
    x = numx.random.random((10,60))
    filters = numx.random.random((3,5,4))
    node = mdp.nodes.Convolution2DNode(filters, input_shape=(3,2))
    with py.test.raises(mdp.NodeException):
        node.execute(x)



from builtins import range
from mdp import graph

def testAddNode():
    # add_node
    g = graph.Graph()
    nnodes = 5
    for i in range(nnodes):
        g.add_node()
    assert len(g.nodes)==nnodes, "Wrong number of nodes, expected: %d, got :%d" % (nnodes, len(g.nodes))
    # add nodes
    g = graph.Graph()
    g.add_nodes(5)
    assert len(g.nodes)==nnodes, "Wrong number of nodes, expected: %d, got :%d" % (nnodes, len(g.nodes))
    g = graph.Graph()
    g.add_nodes([None] * nnodes)
    assert len(g.nodes)==nnodes, "Wrong number of nodes, expected: %d, got :%d" % (nnodes, len(g.nodes))

def testAddEdge():
    g = graph.Graph()
    nnodes = 5
    nds = [g.add_node() for i in range(nnodes)]
    eds = [g.add_edge(nds[i], nds[i+1]) for i in range(nnodes-1)]
    assert len(g.edges)==nnodes-1, "Wrong number of edges, expected: %d, got :%d" % (nnodes-1, len(g.edges))
    # the last nnodes-1 nodes should have in_degree==1,
    # and the first nnodes-1 out_degree==1
    for i in range(nnodes):
        if i>0: assert nds[i].in_degree()==1, "Wrong in_degree, expected: 1, got: %d." % nds[i].in_degree()
        if i<nnodes-1: assert nds[i].out_degree()==1, "Wrong out_degree, expected: 1, got: %d." % nds[i].out_degree()

def testGetEdge():
    g = graph.Graph()
    nds = g.add_nodes(4)
    ed1 = g.add_edge(nds[0],nds[2])
    ed2 = g.add_edge(nds[1],nds[2])
    ed3 = g.add_edge(nds[2],nds[3])
    assert set(nds[2].get_edges_in()) == set([ed1, ed2])
    assert set(nds[2].get_edges_out()) == set([ed3])
    assert set(nds[2].get_edges()) == set([ed1, ed2, ed3])
    assert set(nds[2].get_edges_in(from_=nds[1])) == set([ed2])
    assert set(nds[2].get_edges_out(to_=nds[3])) == set([ed3])
    ed4 = g.add_edge(nds[3],nds[2])
    assert set(nds[2].get_edges(neighbor=nds[3])) == set([ed3,ed4])

def testRemoveEdge():
    g = graph.Graph()
    n0 = g.add_node()
    n1 = g.add_node()
    ed = g.add_edge(n0, n1)
    g.remove_edge(ed)
    assert len(g.edges)==0, "Wrong number of edges, expected: 0, got :%d" % len(g.edges)
    assert n0.out_degree()==0, "Wrong out_degree, expected: 0, got: %d." % n0.out_degree()
    assert n1.in_degree()==0, "Wrong in_degree, expected: 0, got: %d." % n1.in_degree()

def testGetNeighbors():
    g = graph.Graph()
    nodes = g.add_tree( (1,2,(2,3)) )
    g.add_edge(nodes[1],nodes[2][1])
    outnodes = nodes[0].out_neighbors()
    assert nodes[1] in outnodes and nodes[2][0] in outnodes
    innodes = nodes[2][1].in_neighbors()
    assert nodes[2][0] in innodes and nodes[2][0] in innodes
    inoutnodes = nodes[1].neighbors()
    assert nodes[0] in inoutnodes and nodes[2][1] in inoutnodes

def testAddTree():
    g = graph.Graph()
    a = None
    nodes = g.add_tree( (a, a, (a, a, a)) )
    assert len(g.nodes)==5
    assert len(g.edges)==4
    out_deg = graph.recursive_map(lambda x: x.out_degree(), nodes)
    assert out_deg==[2,0,[2,0,0]]
    in_deg = graph.recursive_map(lambda x: x.in_degree(), nodes)
    assert in_deg==[0,1,[1,1,1]]

def testAddFullConnectivity():
    g = graph.Graph()
    layer0 = g.add_nodes(10)
    layer1 = g.add_nodes(5)
    g.add_full_connectivity(layer0, layer1)
    assert len(g.edges)==5*10
    assert [x.out_degree() for x in layer0]==[5]*10
    assert [x.in_degree() for x in layer1]==[10]*5

def testTopologicalSort():
    g = graph.Graph()
    # the values correspond to the number of in-edges
    nds = g.add_tree( (0, 3, 1, 2) )
    g.add_edge(nds[2], nds[1])
    g.add_edge(nds[2], nds[3])
    g.add_edge(nds[3], nds[1])
    data = [x.data for x in g.topological_sort()]
    assert data==[0,1,2,3]
    # make graph cyclic
    g.add_edge(nds[1], nds[0])
    try:
        g.topological_sort()
        raise Exception('Expected graph.GraphTopologicalException.')
    except graph.GraphTopologicalException:
        pass

def testDFS():
    g = graph.Graph()
    nodes = g.add_tree( (1,(2,3),(2,3)) )
    data = [x.data for x in g.dfs(nodes[0])]
    assert data==[1,2,3,2,3]
    # test undirected version
    data = [x.data for x in g.undirected_dfs(nodes[2][1])]
    assert data==[3,2,1,2,3]
    # test node with two ingoing edges
    g = graph.Graph()
    nodes = g.add_tree( (1,2,(2,3)) )
    g.add_edge(nodes[1],nodes[2][1])
    data = [x.data for x in g.dfs(nodes[0])]
    assert data==[1,2,3,2]

def testBFS():
    g = graph.Graph()
    nodes = g.add_tree( (1,(2,3),(2,3)) )
    data = [x.data for x in g.bfs(nodes[0])]
    assert data==[1,2,2,3,3]
    # test undirected version
    data = [x.data for x in g.undirected_bfs(nodes[2][1])]
    assert data==[3,2,1,2,3]
    # test node with two ingoing edges
    g = graph.Graph()
    nodes = g.add_tree( (1,2,(2,3)) )
    g.add_edge(nodes[1],nodes[2][1])
    data = [x.data for x in g.bfs(nodes[0])]
    assert data==[1,2,2,3]

def testConnectedComponents():
    g = graph.Graph()
    nds0 = g.add_tree( (1, 1, 1) )
    nds1 = g.add_tree( (2, 2, 2) )
    comps = g.connected_components()
    comps = graph.recursive_map(lambda x: x.data, comps)
    assert len(comps)==2
    assert comps[0]==[1,1,1]
    assert comps[1]==[2,2,2]
    assert not g.is_weakly_connected()
    # connect graph
    g.add_edge(nds0[0], nds1[0])
    assert g.is_weakly_connected()




"""
Module for MDP Nodes that support parallel training.

This module contains both the parallel base class and some parallel
implementations of MDP nodes. Note that such ParallelNodes are only needed for
training, parallel execution works with any Node that can be pickled.
"""
from builtins import str

# WARNING: There is a problem with unpickled arrays in NumPy < 1.1.x, see
# http://projects.scipy.org/scipy/numpy/ticket/551
# To circumvent this, you can use a copy() of all unpickled arrays.

import inspect

import mdp
from mdp import numx


class NotForkableParallelException(mdp.NodeException):
    """Exception to signal that a fork is not possible.

    This exception is can be safely used and should be caught inside the
    ParallelFlow or the Scheduler.
    """
    pass


class JoinParallelException(mdp.NodeException):
    """Exception for errors when joining parallel nodes."""
    pass


class ParallelExtensionNode(mdp.ExtensionNode, mdp.Node):
    """Base class for parallel trainable MDP nodes.

    With the fork method new node instances are created which can then be
    trained. With the join method the trained instances are then merged back
    into a single node instance.

    This class defines default methods which raise a
    TrainingPhaseNotParallelException exception.
    """

    extension_name = "parallel"

    # TODO: allow that forked nodes are not forkable themselves,
    #    and are not joinable either
    #    this implies that caching does not work for these
    
    def fork(self):
        """Return a new instance of this node class for remote training.

        This is a template method, the actual forking should be implemented in
        _fork.

        The forked node should be a ParallelNode of the same class as well,
        thus allowing recursive forking and joining.
        """
        return self._fork()

    def join(self, forked_node):
        """Absorb the trained node from a fork into this parent node.

        This is a template method, the actual joining should be implemented in
        _join.
        """
        # Warning: Use the properties / setters here. Otherwise we get problems
        #    in certain situations (e.g., for FlowNode).
        if self.dtype is None:
            self.dtype = forked_node.dtype
        if self.input_dim is None:
            self.input_dim = forked_node.input_dim
        if self.output_dim is None:
            self.output_dim = forked_node.output_dim
        if forked_node._train_phase_started and not self._train_phase_started:
            self._train_phase_started = True
        self._join(forked_node)

    ## overwrite these methods ##

    def _fork(self):
        """Hook method for forking with default implementation.

        Overwrite this method for nodes that can be parallelized.
        You can use _default_fork, if that is compatible with your node class,
        typically the hard part is the joining.
        """
        raise NotForkableParallelException("fork is not implemented " +
                                           "by this node (%s)" %
                                           str(self.__class__))

    def _join(self, forked_node):
        """Hook method for joining, to be overridden."""
        raise JoinParallelException("join is not implemented " +
                                    "by this node (%s)" %
                                    str(self.__class__))
    
    @staticmethod
    def use_execute_fork():
        """Return True if node requires a fork / join even during execution.
        
        The default output is False, overwrite this method if required.
        
        Note that the same fork and join methods are used as during training,
        so the distinction must be implemented in the custom _fork and _join
        methods.
        """
        return False
    
    ## helper methods ##

    def _default_fork(self):
        """Default implementation of _fork.

        It uses introspection to determine the init kwargs and tries to fill
        them with attributes. These kwargs are then used to instanciate
        self.__class__ to create the fork instance.

        So you can use this method if all the required keys are also public
        attributes or have a single underscore in front.
        
        There are two reasons why this method does not simply replace _fork
        of ParallelExtensionNode (plus removing Node from the
        inheritance list):
        - If a node is not parallelized _fork raises an exception, as do nodes
            which can not fork due to some other reasons. Without this bahavior
            of _fork we would have to check with hasattr first if fork is
            present, adding more complexity at other places (mostly in
            container nodes).
        - This is a safeguard forcing users to think a little instead of
            relying on the inherited (but possibly incompatible)
            default implementation.
        """
        args, varargs, varkw, defaults = inspect.getargspec(self.__init__)
        args.remove("self")
        if defaults:
            non_default_keys = args[:-len(defaults)]
        else:
            non_default_keys = []
        kwargs = dict((key, getattr(self, key))
                      for key in args if hasattr(self, key))
        # look for the key with an underscore in front
        for key in kwargs:
            args.remove(key)
        under_kwargs = dict((key, getattr(self, '_' + key))
                            for key in args if hasattr(self, '_' + key))
        for key in under_kwargs:
            args.remove(key)
        kwargs.update(under_kwargs)
        # check that all the keys without default arguments are covered
        if non_default_keys:
            missing_defaults = set(non_default_keys) & set(args)
            if missing_defaults:
                err = ("could not find attributes for init arguments %s" %
                       str(missing_defaults))
                raise NotForkableParallelException(err)
        # create new instance
        return self.__class__(**kwargs)
    
    @staticmethod
    def _join_covariance(cov, forked_cov):
        """Helper method to join two CovarianceMatrix instances.
        
        cov -- Instance of CovarianceMatrix, to which the forked_cov instance
            is aded in-place.
        """
        cov._cov_mtx += forked_cov._cov_mtx
        cov._avg += forked_cov._avg
        cov._tlen += forked_cov._tlen
        

## MDP parallel node implementations ##

class ParallelPCANode(ParallelExtensionNode, mdp.nodes.PCANode):
    """Parallel version of MDP PCA node."""

    def _fork(self):
        return self._default_fork()

    def _join(self, forked_node):
        """Combine the covariance matrices."""
        if self._cov_mtx._cov_mtx is None:
            self.set_dtype(self._cov_mtx._dtype)
            self._cov_mtx = forked_node._cov_mtx
        else:
            self._join_covariance(self._cov_mtx, forked_node._cov_mtx)


class ParallelSFANode(ParallelExtensionNode, mdp.nodes.SFANode):
    """Parallel version of MDP SFA node."""

    def _fork(self):
        return self._default_fork()

    def _join(self, forked_node):
        """Combine the covariance matrices."""
        if self._cov_mtx._cov_mtx is None:
            self.set_dtype(forked_node._cov_mtx._dtype)
            self._cov_mtx = forked_node._cov_mtx
            self._dcov_mtx = forked_node._dcov_mtx
        else:
            self._join_covariance(self._cov_mtx, forked_node._cov_mtx)
            self._join_covariance(self._dcov_mtx, forked_node._dcov_mtx)


class ParallelFDANode(ParallelExtensionNode, mdp.nodes.FDANode):

    def _fork(self):
        if self.get_current_train_phase() == 1:
            forked_node = self.copy()
            # reset the variables that might contain data from this train phase
            forked_node._S_W = None
            forked_node._allcov = mdp.utils.CovarianceMatrix(dtype=self.dtype)
        else:
            forked_node = self._default_fork()
        return forked_node

    def _join(self, forked_node):
        if self.get_current_train_phase() == 1:
            if forked_node.get_current_train_phase() != 1:
                msg = ("This node is in training phase 1, but the forked node "
                       "is not.")
                raise NotForkableParallelException(msg)
            if self._S_W is None:
                self.set_dtype(forked_node._allcov._dtype)
                self._allcov = forked_node._allcov
                self._S_W = forked_node._S_W
            else:
                self._join_covariance(self._allcov, forked_node._allcov)
                self._S_W += forked_node._S_W
        else:
            for lbl in forked_node.means:
                if lbl in self.means:
                    self.means[lbl] += forked_node.means[lbl]
                    self.tlens[lbl] += forked_node.tlens[lbl]
                else:
                    self.means[lbl] = forked_node.means[lbl]
                    self.tlens[lbl] = forked_node.tlens[lbl]


class ParallelHistogramNode(ParallelExtensionNode, mdp.nodes.HistogramNode):
    """Parallel version of the HistogramNode."""

    def _fork(self):
        return self._default_fork()

    def _join(self, forked_node):
        if (self.data_hist is not None) and (forked_node.data_hist is not None):
            self.data_hist = numx.concatenate([self.data_hist,
                                            forked_node.data_hist])
        elif forked_node.data_hist != None:
            self.data_hist = forked_node.data_hist



"""
Process based scheduler for distribution across multiple CPU cores.
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range

# TODO: use a queue instead of sleep?
#    http://docs.python.org/library/queue.html

# TODO: use shared memory for data numpy arrays, but this also requires the
#    use of multiprocessing since the ctype objects can't be pickled

# TODO: only return result when get_results is called,
#    this sends a special request to the processes to send their data,
#    we would have to add support for this to the callable,
#    might get too complicated

# TODO: leverage process forks on unix systems,
#    might be very efficient due to copy-on-write, see
#    http://gael-varoquaux.info/blog/?p=119
#    http://www.ibm.com/developerworks/aix/library/au-multiprocessing/

import sys
import os
import pickle as pickle
import threading
import subprocess
import time
import traceback
import warnings

if __name__ == "__main__":
    # try to make sure that mdp can be imported by adding it to sys.path
    mdp_path = os.path.realpath(__file__)
    mdp_index = mdp_path.rfind("mdp")
    if mdp_index:
        mdp_path = mdp_path[:mdp_index-1]
        # mdp path goes after sys.path
        sys.path.append(mdp_path)
    # shut off warnings of any kinds
    warnings.filterwarnings("ignore", ".*")

import mdp
from mdp.parallel import Scheduler, cpu_count

SLEEP_TIME = 0.1  # time spend sleeping when waiting for a free process


class ProcessScheduler(Scheduler):
    """Scheduler that distributes the task to multiple processes.

    The subprocess module is used to start the requested number of processes.
    The execution of each task is internally managed by dedicated thread.

    This scheduler should work on all platforms (at least on Linux,
    Windows XP and Vista).
    """

    def __init__(self, result_container=None, verbose=False, n_processes=1,
                 source_paths=None, python_executable=None,
                 cache_callable=True):
        """Initialize the scheduler and start the slave processes.

        result_container -- ResultContainer used to store the results.
        verbose -- Set to True to get progress reports from the scheduler
            (default value is False).
        n_processes -- Number of processes used in parallel. If None (default)
            then the number of detected CPU cores is used.
        source_paths -- List of paths that are added to sys.path in
            the processes to make the task unpickling work. A single path
            instead of a list is also accepted.
            If None (default value) then source_paths is set to sys.path.
            To prevent this you can specify an empty list.
        python_executable -- Python executable that is used for the processes.
            The default value is None, in which case sys.executable will be
            used.
        cache_callable -- Cache the task objects in the processes (default
            is True). Disabling caching can reduce the memory usage, but will
            generally be less efficient since the task_callable has to be
            pickled each time.
        """
        super(ProcessScheduler, self).__init__(
                                        result_container=result_container,
                                        verbose=verbose)
        if n_processes:
            self._n_processes = n_processes
        else:
            self._n_processes = cpu_count()
        self._cache_callable = cache_callable
        if python_executable is None:
            python_executable = sys.executable
        # get the location of this module to start the processes
        module_path = os.path.dirname(mdp.__file__)
        module_file = os.path.join(module_path, "parallel",
                                   "process_schedule.py")
        # Note: -u argument is important on Windows to set stdout to binary
        #    mode. Otherwise you might get a strange error message for
        #    copy_reg.
        process_args = [python_executable, "-u", module_file]
        process_args.append(str(self._cache_callable))
        if isinstance(source_paths, str):
            source_paths = [source_paths]
        if source_paths is None:
            source_paths = sys.path
        process_args += source_paths
        # list of processes not in use, start the processes now
        self._free_processes = [subprocess.Popen(args=process_args,
                                                 stdout=subprocess.PIPE,
                                                 stdin=subprocess.PIPE)
                                for _ in range(self._n_processes)]
        # tag each process with its cached callable task_index,
        # this is compared with _last_callable_index to check if the cached
        # task_callable is still up to date
        for process in self._free_processes:
            process._callable_index = -1
        if self.verbose:
            print ("scheduler initialized with %d processes" %
                   self._n_processes)

    def _shutdown(self):
        """Shut down the slave processes.

        If a process is still running a task then an exception is raised.
        """
        self._lock.acquire()
        if len(self._free_processes) < self._n_processes:
            raise Exception("some slave process is still working")
        for process in self._free_processes:
            pickle.dump("EXIT", process.stdin)
            process.stdin.flush()
        self._lock.release()
        if self.verbose:
            print("scheduler shutdown")

    def _process_task(self, data, task_callable, task_index):
        """Add a task, if possible without blocking.

        It blocks when the system is not able to start a new thread
        or when the processes are all in use.
        """
        task_started = False
        while not task_started:
            if not len(self._free_processes):
                # release lock for other threads and wait
                self._lock.release()
                time.sleep(SLEEP_TIME)
                self._lock.acquire()
            else:
                try:
                    process = self._free_processes.pop()
                    self._lock.release()
                    thread = threading.Thread(target=self._task_thread,
                                              args=(process, data,
                                                    task_callable, task_index))
                    thread.start()
                    task_started = True
                except thread.error:
                    if self.verbose:
                        print ("unable to create new task thread,"
                               " waiting 2 seconds...")
                    time.sleep(2)

    def _task_thread(self, process, data, task_callable, task_index):
        """Thread function which cares for a single task.

        The task is pushed to the process via stdin, then we wait for the
        result on stdout, pass the result to the result container, free
        the process and exit.
        """
        try:
            if self._cache_callable:
                # check if the cached callable is up to date
                if process._callable_index < self._last_callable_index:
                    process._callable_index = self._last_callable_index
                else:
                    task_callable = None
            # push the task to the process
            pickle.dump((data, task_callable, task_index),
                        process.stdin, protocol=-1)
            process.stdin.flush()
            # wait for result to arrive
            result = pickle.load(process.stdout)
        except:
            traceback.print_exc()
            self._free_processes.append(process)
            sys.exit("failed to execute task %d in process:" % task_index)
        # store the result and clean up
        self._store_result(result, task_index)
        self._free_processes.append(process)


def _process_run(cache_callable=True):
    """Run this function in a worker process to receive and run tasks.

    It waits for tasks on stdin, and sends the results back via stdout.
    """
    # use sys.stdout only for pickled objects, everything else goes to stderr
    # NOTE: .buffer is the binary mode interface for stdin and out in py3k
    try:
        pickle_out = sys.stdout.buffer
    except AttributeError:
        pickle_out = sys.stdout
    try:
        pickle_in = sys.stdin.buffer
    except AttributeError:
        pickle_in = sys.stdin

    sys.stdout = sys.stderr
    exit_loop = False
    last_callable = None  # cached callable
    while not exit_loop:
        task = None
        try:
            # wait for task to arrive
            task = pickle.load(pickle_in)
            if task == "EXIT":
                exit_loop = True
            else:
                data, task_callable, task_index = task
                if task_callable is None:
                    if last_callable is None:
                        err = ("No callable was provided and no cached "
                               "callable is available.")
                        raise Exception(err)
                    task_callable = last_callable.fork()
                elif cache_callable:
                    # store callable in cache
                    last_callable = task_callable
                    task_callable.setup_environment()
                    task_callable = task_callable.fork()
                else:
                    task_callable.setup_environment()
                result = task_callable(data)
                del task_callable  # free memory
                pickle.dump(result, pickle_out, protocol=-1)
                pickle_out.flush()
        except Exception as exception:
            # return the exception instead of the result
            if task is None:
                print("unpickling a task caused an exception in a process:")
            else:
                print("task %d caused exception in process:" % task[2])
            print(exception)
            traceback.print_exc()
            sys.stdout.flush()
            sys.exit()

if __name__ == "__main__":
    # first argument is cache_callable flag
    cache_callable = sys.argv[1] == "True"

    if len(sys.argv) > 2:
        # remaining arguments are code paths,
        # put them in front so that they take precedence over PYTHONPATH
        new_paths = [sys_arg for sys_arg in sys.argv[2:]
                     if sys_arg not in sys.path]
        sys.path = new_paths + sys.path
    _process_run(cache_callable=cache_callable)



from __future__ import division
from past.utils import old_div
import py.test
from ._tools import *

def test_FDANode():
    mean1 = [0., 2.]
    mean2 = [0., -2.]
    std_ = numx.array([1., 0.2])
    npoints = 50000
    rot = 45

    # input data: two distinct gaussians rotated by 45 deg
    def distr(size):
        return normal(0, 1., size=(size)) * std_

    x1 = distr((npoints,2)) + mean1
    utils.rotate(x1, rot, units='degrees')
    x2 = distr((npoints,2)) + mean2
    utils.rotate(x2, rot, units='degrees')
    x = numx.concatenate((x1, x2), axis=0)

    # labels
    cl1 = numx.ones((x1.shape[0],), dtype='d')
    cl2 = 2.*numx.ones((x2.shape[0],), dtype='d')
    classes = numx.concatenate((cl1, cl2))

    # shuffle the data
    perm_idx = numx_rand.permutation(classes.shape[0])
    x = numx.take(x, perm_idx, axis=0)

    classes = numx.take(classes, perm_idx)

    flow = mdp.Flow([mdp.nodes.FDANode()])
    py.test.raises(mdp.TrainingException,
                   flow[0].train, x, numx.ones((2,)))

    flow.train([[(x, classes)]])
    fda_node = flow[0]

    assert fda_node.tlens[1] == npoints
    assert fda_node.tlens[2] == npoints

    m1 = numx.array([mean1])
    m2 = numx.array([mean2])
    utils.rotate(m1, rot, units='degrees')
    utils.rotate(m2, rot, units='degrees')
    assert_array_almost_equal(fda_node.means[1], m1, 2)
    assert_array_almost_equal(fda_node.means[2], m2, 2)

    y = flow.execute(x)
    assert_array_almost_equal(mean(y, axis=0), [0., 0.], decimal-2)
    assert_array_almost_equal(std(y, axis=0), [1., 1.], decimal-2)
    assert_almost_equal(mult(y[:,0], y[:,1].T), 0., decimal-2)

    v1 = old_div(fda_node.v[:,0],fda_node.v[0,0])
    assert_array_almost_equal(v1, [1., -1.], 2)
    v1 = old_div(fda_node.v[:,1],fda_node.v[0,1])
    assert_array_almost_equal(v1, [1., 1.], 2)



from builtins import range
from builtins import object
import mdp
import bimdp

from mdp import numx, numx_rand

class TestGradientExtension(object):

    def test_sfa_gradient(self):
        """Test gradient for combination of SFA nodes."""
        sfa_node1 = bimdp.nodes.SFABiNode(output_dim=8)
        sfa_node2 = bimdp.nodes.SFABiNode(output_dim=7)
        sfa_node3 = bimdp.nodes.SFABiNode(output_dim=5)
        flow = sfa_node1 + sfa_node2 + sfa_node3
        x = numx_rand.random((300, 10))
        flow.train(x)
        x = numx_rand.random((2, 10))
        mdp.activate_extension("gradient")
        try:
            flow.execute(x, {"method": "gradient"})
        finally:
            mdp.deactivate_extension("gradient")

    def test_gradient_product(self):
        """Test that the product of gradients is calculated correctly."""
        sfa_node1 = bimdp.nodes.SFABiNode(output_dim=5)
        sfa_node2 = bimdp.nodes.SFABiNode(output_dim=3)
        flow = sfa_node1 + sfa_node2
        x = numx_rand.random((300, 10))
        flow.train(x)
        mdp.activate_extension("gradient")
        try:
            x1 = numx_rand.random((2, 10))
            x2, msg = sfa_node1.execute(x1, {"method": "gradient"})
            grad1 = msg["grad"]
            _, msg = sfa_node2.execute(x2, {"method": "gradient"})
            grad2 = msg["grad"]
            grad12 = flow.execute(x1, {"method": "gradient"})[1]["grad"]
            # use a different way to calculate the product of the gradients,
            # this method is too memory intensive for large data
            ref_grad = numx.sum(grad2[:,:,numx.newaxis,:] *
                             numx.transpose(grad1[:,numx.newaxis,:,:], (0,1,3,2)),
                             axis=3)
            assert numx.amax(abs(ref_grad - grad12)) < 1E-9
        finally:
            mdp.deactivate_extension("gradient")

    def test_quadexpan_gradient1(self):
        """Test validity of gradient for QuadraticExpansionBiNode."""
        node = mdp.nodes.QuadraticExpansionNode()
        x = numx.array([[1, 3, 4]])
        node.execute(x)
        mdp.activate_extension("gradient")
        try:
            result = node._gradient(x)
            grad = result[1]["grad"]
            reference = numx.array(
                [[[ 1, 0, 0],   # x1
                  [ 0, 1, 0],   # x2
                  [ 0, 0, 1],   # x3
                  [ 2, 0, 0],   # x1x1
                  [ 3, 1, 0],   # x1x2
                  [ 4, 0, 1],   # x1x3
                  [ 0, 6, 0],   # x2x2
                  [ 0, 4, 3],   # x2x3
                  [ 0, 0, 8]]]) # x3x3
            assert numx.all(grad == reference)
        finally:
            mdp.deactivate_extension("gradient")

    def test_quadexpan_gradient2(self):
        """Test gradient with multiple data points."""
        node = mdp.nodes.QuadraticExpansionNode()
        x = numx_rand.random((3,5))
        node.execute(x)
        mdp.activate_extension("gradient")
        try:
            result = node._gradient(x)
            gradient = result[1]["grad"]
            assert gradient.shape == (3,20,5)
        finally:
            mdp.deactivate_extension("gradient")

    def test_sfa2_gradient(self):
        sfa2_node1 = bimdp.nodes.SFA2BiNode(output_dim=5)
        sfa2_node2 = bimdp.nodes.SFA2BiNode(output_dim=3)
        flow = sfa2_node1 + sfa2_node2
        x = numx_rand.random((300, 6))
        flow.train(x)
        x = numx_rand.random((2, 6))
        mdp.activate_extension("gradient")
        try:
            flow.execute(x, {"method": "gradient"})
        finally:
            mdp.deactivate_extension("gradient")

    def test_sfa2_gradient2(self):
        def _alt_sfa2_grad(self, x):
            """Reference grad method based on quadratic forms."""
            # note that the H and f arrays are cached in the node and remain even
            # after the extension has been deactivated
            if not hasattr(self, "__gradient_Hs"):
                quad_forms = [self.get_quadratic_form(i)
                              for i in range(self.output_dim)]
                self.__gradient_Hs = numx.vstack((quad_form.H[numx.newaxis]
                                                for quad_form in quad_forms))
                self.__gradient_fs = numx.vstack((quad_form.f[numx.newaxis]
                                                for quad_form in quad_forms))
            grad = (numx.dot(x, self.__gradient_Hs) +
                        numx.repeat(self.__gradient_fs[numx.newaxis,:,:],
                                  len(x), axis=0))
            return grad
        sfa2_node = bimdp.nodes.SFA2BiNode(output_dim=3)
        x = numx_rand.random((300, 6))
        sfa2_node.train(x)
        sfa2_node.stop_training()
        x = numx_rand.random((2, 6))
        mdp.activate_extension("gradient")
        try:
            result1 = sfa2_node.execute(x, {"method": "gradient"})
            grad1 = result1[1]["grad"]
            grad2 = _alt_sfa2_grad(sfa2_node, x)
            assert numx.amax(abs(grad1 - grad2)) < 1E-9
        finally:
            mdp.deactivate_extension("gradient")

    def test_layer_gradient(self):
        """Test gradient for a simple layer."""
        node1 = mdp.nodes.SFA2Node(input_dim=4, output_dim=3)
        node2 = mdp.nodes.SFANode(input_dim=6, output_dim=2)
        layer = mdp.hinet.Layer([node1, node2])
        x = numx_rand.random((100,10))
        layer.train(x)
        layer.stop_training()
        mdp.activate_extension("gradient")
        try:
            x = numx_rand.random((7,10))
            result = layer._gradient(x)
            grad = result[1]["grad"]
            # get reference result
            grad1 = node1._gradient(x[:, :node1.input_dim])[1]["grad"]
            grad2 = node2._gradient(x[:, node1.input_dim:])[1]["grad"]
            ref_grad = numx.zeros(((7,5,10)))
            ref_grad[:, :node1.output_dim, :node1.input_dim] = grad1
            ref_grad[:, node1.output_dim:, node1.input_dim:] = grad2
            assert numx.all(grad == ref_grad)
        finally:
            mdp.deactivate_extension("gradient")

    def test_clonebilayer_gradient(self):
        """Test gradient for a simple layer."""
        layer = bimdp.hinet.CloneBiLayer(
                            bimdp.nodes.SFA2BiNode(input_dim=5, output_dim=2),
                            n_nodes=3)
        x = numx_rand.random((100,15))
        layer.train(x)
        layer.stop_training()
        mdp.activate_extension("gradient")
        try:
            x = numx_rand.random((7,15))
            result = layer._gradient(x)
            grad = result[1]["grad"]
            assert grad.shape == (7,6,15)
        finally:
            mdp.deactivate_extension("gradient")

    def test_switchboard_gradient1(self):
        """Test that gradient is correct for a tiny switchboard."""
        sboard = mdp.hinet.Switchboard(input_dim=4, connections=[2,0])
        x = numx_rand.random((2,4))
        mdp.activate_extension("gradient")
        try:
            result = sboard._gradient(x)
            grad = result[1]["grad"]
            ref_grad = numx.array([[[0,0,1,0], [1,0,0,0]],
                                 [[0,0,1,0], [1,0,0,0]]], dtype=grad.dtype)
            assert numx.all(grad == ref_grad)
        finally:
            mdp.deactivate_extension("gradient")

    def test_switchboard_gradient2(self):
        """Test gradient for a larger switchboard."""
        dim = 100
        connections = [int(i) for i in numx.random.random((dim,)) * (dim-1)]
        sboard = mdp.hinet.Switchboard(input_dim=dim, connections=connections)
        x = numx.random.random((10, dim))
        # assume a 5-dimensional gradient at this stage
        grad = numx.random.random((10, dim, 5))
        # original reference implementation
        def _switchboard_grad(self, x):
            grad = numx.zeros((self.output_dim, self.input_dim))
            grad[list(range(self.output_dim)), self.connections] = 1
            return numx.tile(grad, (len(x), 1, 1))
        with mdp.extension("gradient"):
            result = sboard._gradient(x, grad)
            ext_grad = result[1]["grad"]
            tmp_grad = _switchboard_grad(sboard, x)
            ref_grad = numx.asarray([numx.dot(tmp_grad[i], grad[i])
                                     for i in range(len(tmp_grad))])
        assert numx.all(ext_grad == ref_grad)

    def test_network_gradient(self):
        """Test gradient for a small SFA network."""
        sfa_node = bimdp.nodes.SFABiNode(input_dim=4*4, output_dim=5)
        switchboard = bimdp.hinet.Rectangular2dBiSwitchboard(
                                                  in_channels_xy=8,
                                                  field_channels_xy=4,
                                                  field_spacing_xy=2)
        flownode = bimdp.hinet.BiFlowNode(bimdp.BiFlow([sfa_node]))
        sfa_layer = bimdp.hinet.CloneBiLayer(flownode,
                                             switchboard.output_channels)
        flow = bimdp.BiFlow([switchboard, sfa_layer])
        train_gen = [numx_rand.random((10, switchboard.input_dim))
                     for _ in range(3)]
        flow.train([None, train_gen])
        # now can test the gradient
        mdp.activate_extension("gradient")
        try:
            x = numx_rand.random((3, switchboard.input_dim))
            result = flow(x, {"method": "gradient"})
            grad = result[1]["grad"]
            assert grad.shape == (3, sfa_layer.output_dim,
                                  switchboard.input_dim)
        finally:
            mdp.deactivate_extension("gradient")



from ._tools import *


def testNoiseNode():
    def bogus_noise(mean, size=None):
        return numx.ones(size)*mean

    node = mdp.nodes.NoiseNode(bogus_noise, (1.,))
    out = node.execute(numx.zeros((100,10),'d'))
    assert_array_equal(out, numx.ones((100,10),'d'))
    node = mdp.nodes.NoiseNode(bogus_noise, (1.,), 'multiplicative')
    out = node.execute(numx.zeros((100,10),'d'))
    assert_array_equal(out, numx.zeros((100,10),'d'))

def testNormalNoiseNode():
    node = mdp.nodes.NormalNoiseNode(noise_args=(2.1, 0.1))
    x = numx.zeros((20000, 10))
    y = node.execute(x)
    assert numx.allclose(y.mean(0), 2.1, atol=1e-02)
    assert numx.allclose(y.std(0), 0.1, atol=1e-02)

def testNoiseNodePickling():
    node = mdp.nodes.NoiseNode()
    node.copy()
    node.save(None)



"""These are some benchmark functions for MDP.
"""
from __future__ import print_function
from builtins import str
from builtins import range

import mdp
#from mdp.utils import symeig
from mdp.utils import matmult as mult

numx = mdp.numx
numx_rand = mdp.numx_rand
numx_fft = mdp.numx_fft

####### benchmark function

def matmult_c_MDP_benchmark(dim):
    """    This benchmark multiplies two contiguous matrices using the
    MDP internal matrix multiplication routine.
    First argument matrix dimensionality"""
    a = numx_rand.random((dim,dim))
    b = numx_rand.random((dim,dim))
    mult(a,b)

def matmult_c_scipy_benchmark(dim):
    """    This benchmark multiplies two contiguous matrices using the
    scipy internal matrix multiplication routine.
    First argument matrix dimensionality"""
    a = numx_rand.random((dim,dim))
    b = numx_rand.random((dim,dim))
    numx.dot(a,b)

def matmult_n_MDP_benchmark(dim):
    """    This benchmark multiplies two non-contiguous matrices using the
    MDP internal matrix multiplication routine.
    First argument matrix dimensionality"""
    a = numx_rand.random((dim,dim)).T
    b = numx_rand.random((dim,dim)).T
    mult(a,b)

def matmult_n_scipy_benchmark(dim):
    """    This benchmark multiplies two non-contiguous matrices using the
    scipy internal matrix multiplication routine.
    First argument matrix dimensionality"""
    a = numx_rand.random((dim,dim)).T
    b = numx_rand.random((dim,dim)).T
    numx.dot(a,b)

def matmult_cn_MDP_benchmark(dim):
    """    This benchmark multiplies a contiguous matrix with a
    non-contiguous matrix using the MDP internal matrix multiplication
    routine.
    First argument matrix dimensionality"""
    a = numx_rand.random((dim,dim)).T
    b = numx_rand.random((dim,dim))
    mult(a,b)

def matmult_cn_scipy_benchmark(dim):
    """    This benchmark multiplies a contiguous matrix with a
    non-contiguous matrix using the scipy internal matrix multiplication
    routine.
    First argument matrix dimensionality"""
    a = numx_rand.random((dim,dim)).T
    b = numx_rand.random((dim,dim))
    numx.dot(a,b)

def quadratic_expansion_benchmark(dim, len, times):
    """    This benchmark expands random data of shape (len, dim)
    'times' times.
    Arguments: (dim,len,times)."""
    a = numx_rand.random((len,dim))
    qnode = mdp.nodes.QuadraticExpansionNode()
    for i in range(times):
        qnode(a)

def polynomial_expansion_benchmark(dim, len, degree, times):
    """    This benchmark expands random data of shape (len, dim)
    'times' times in the space of polynomials of degree 'degree'.
    Arguments: (dim,len,degree,times)."""
    numx_rand.seed(4253529)
    a = numx_rand.random((len,dim))
    pnode = mdp.nodes.PolynomialExpansionNode(degree)
    for i in range(times):
        pnode(a)

# ISFA benchmark

def _tobias_mix(src):
    mix = src.copy()
    mix[:,0]=(src[:,1]+3*src[:,0]+6)*numx.cos(1.5*numx.pi*src[:,0])
    mix[:,1]=(src[:,1]+3*src[:,0]+6)*numx.sin(1.5*numx.pi*src[:,0])
    return mix

def _get_random_slow_sources(nsrc, distr_fun):
    # nsrc: number of sources
    # distr_fun: random numbers function

    src = distr_fun(size=(50000, nsrc))
    fsrc = numx_fft.rfft(src, axis=0)
    # enforce different time scales
    for i in range(nsrc):
        fsrc[5000+(i+1)*1000:,i] = 0.
    src = numx_fft.irfft(fsrc,axis=0)
    return src

def isfa_spiral_benchmark():
    """    Apply ISFA to twisted data."""
    numx_rand.seed(116599099)
    # create independent sources
    src = _get_random_slow_sources(2, numx_rand.laplace)
    # subtract mean and rescale between -1 and 1
    src -= src.mean(axis=0)
    src /= abs(src).max()
    # apply nonlinear "twist" transformation
    exp_src = _tobias_mix(src)
    # train
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(5),
                     mdp.nodes.SFANode(),
                     mdp.nodes.ISFANode(lags=30, whitened=False,
                                        sfa_ica_coeff=[1.,300.],
                                        eps_contrast=1e-5,
                                        output_dim=2, verbose=False)])
    flow.train(exp_src)

def sfa_benchmark():
    """    Apply SFA to twisted data."""
    numx_rand.seed(424507)
    # create independent sources
    nsrc = 15
    src = _get_random_slow_sources(nsrc, numx_rand.normal)
    src = src[:5000,:]
    src = mult(src, numx_rand.uniform(size=(nsrc, nsrc))) \
          + numx_rand.uniform(size=nsrc)
    # train
    flow = mdp.Flow([mdp.nodes.PolynomialExpansionNode(3),
                     mdp.nodes.PCANode(output_dim = 100),
                     mdp.nodes.SFANode(output_dim = 30)])
    #src = src.reshape(1000,5,nsrc)
    flow.train([None, [src], [src]])

#### benchmark tools

# function used to measure time
import time
TIMEFUNC = time.time

def timeit(func,*args,**kwargs):
    """Return function execution time in 1/100ths of a second."""
    tstart = TIMEFUNC()
    func(*args,**kwargs)
    return (TIMEFUNC()-tstart)*100.

def _random_seed():
    import sys
    seed = int(numx_rand.randint(2**31-1))
    numx_rand.seed(seed)
    sys.stderr.write("Random Seed: " + str(seed)+'\n')

def run_benchmarks(bench_funcs, time_digits=15):

    results_str = '| %%s | %%%d.2f |' % time_digits
    label_str = '| %%s | %s |' % 'Time (sec/100)'.center(time_digits)

    tstart = TIMEFUNC()

    # loop over all benchmarks functions
    for func, args_list in bench_funcs:
        # number of combinations of arguments(cases)
        ncases = len(args_list)
        funcname = func.__name__[:-10]

        # loop over all cases
        for i in range(ncases):
            args = args_list[i]

            # format description string
            descr = funcname + str(tuple(args))
            if i==0:
                # print summary table header
                descrlen = len(descr)+6
                results_strlen = time_digits+descrlen+7
                print('\nTiming results (%s, %d cases):' % (funcname, ncases))
                print(func.__doc__)
                print('+'+'-'*(results_strlen-2)+'+')
                print(label_str % 'Description'.center(descrlen))
                print('+'+'-'*(results_strlen-2)+'+')

            # execute function
            t = timeit(func, *args)

            # print summary table entry
            print(results_str % (descr.center(descrlen), t))

        # print summary table tail
        print('+'+'-'*(results_strlen-2)+'+')

    print('\nTotal running time:', (TIMEFUNC()-tstart)*100.)

####### /benchmark function

POLY_EXP_ARGS = [(2**i, 100, j, 200) for j in range(2,5) for i in range(2,4)]

#if mdp.numx_description in ['symeig', 'scipy', 'numpy']:
#    MUL_MTX_DIMS = [[2**i] for i in xrange(4,11)]
#    # list of (benchmark function, list of arguments)
#    BENCH_FUNCS = [(matmult_c_MDP_benchmark, MUL_MTX_DIMS),
#                   (matmult_c_scipy_benchmark, MUL_MTX_DIMS),
#                   (matmult_n_MDP_benchmark, MUL_MTX_DIMS),
#                   (matmult_n_scipy_benchmark, MUL_MTX_DIMS),
#                   (matmult_cn_MDP_benchmark, MUL_MTX_DIMS),
#                   (matmult_cn_scipy_benchmark, MUL_MTX_DIMS),
#                   (polynomial_expansion_benchmark, POLY_EXP_ARGS)]
#else:
#    BENCH_FUNCS = [(polynomial_expansion_benchmark, POLY_EXP_ARGS)]
BENCH_FUNCS = [(polynomial_expansion_benchmark, POLY_EXP_ARGS),
               (isfa_spiral_benchmark, [[]]),
               (sfa_benchmark, [[]])]

def get_benchmarks():
    return BENCH_FUNCS

if __name__ == "__main__":
    print("Running benchmarks: ")
    run_benchmarks(get_benchmarks())



__docformat__ = "restructuredtext en"

import mdp

numx = mdp.numx

class FDANode(mdp.Node):
    """Perform a (generalized) Fisher Discriminant Analysis of its
    input. It is a supervised node that implements FDA using a
    generalized eigenvalue approach.

    FDANode has two training phases and is supervised so make sure to
    pay attention to the following points when you train it:

    - call the ``train`` method with *two* arguments: the input data
      and the labels (see the doc string of the ``train`` method for details).

    - if you are training the node by hand, call the ``train`` method twice.

    - if you are training the node using a flow (recommended), the
      only argument to ``Flow.train`` must be a list of
      ``(data_point, label)`` tuples or an iterator returning lists of
      such tuples, *not* a generator.  The ``Flow.train`` function can be
      called just once as usual, since it takes care of *rewinding* the iterator
      to perform the second training step.

    More information on Fisher Discriminant Analysis can be found for
    example in C. Bishop, Neural Networks for Pattern Recognition,
    Oxford Press, pp. 105-112.

    **Internal variables of interest**

      ``self.avg``
          Mean of the input data (available after training)

      ``self.v``
          Transposed of the projection matrix, so that
          ``output = dot(input-self.avg, self.v)`` (available after training).
    """

    def _get_train_seq(self):
        return [(self._train_means, self._stop_means),
                (self._train_fda, self._stop_fda)]

    def __init__(self, input_dim=None, output_dim=None, dtype=None):
        super(FDANode, self).__init__(input_dim, output_dim, dtype)
        # mean in-class covariance matrix times number of data points
        # is deleted after training
        self._S_W = None
        # covariance matrix of the full data distribution
        self._allcov = mdp.utils.CovarianceMatrix(dtype=self.dtype)
        self.means = {}  # maps class labels to the class means
        self.tlens = {}  # maps class labels to number of training points
        self.v = None  # transposed of the projection matrix
        self.avg = None  # mean of the input data

    def _check_train_args(self, x, labels):
        if (isinstance(labels, (list, tuple, numx.ndarray)) and
            len(labels) != x.shape[0]):
            msg = ("The number of labels should be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)

    # Training step 1: compute mean and number of elements in each class

    def _train_means(self, x, labels):
        """Gather data to compute the means and number of elements."""
        if isinstance(labels, (list, tuple, numx.ndarray)):
            labels_ = numx.asarray(labels)
            for label in set(labels_):
                # group for class
                x_label = numx.compress(labels_==label, x, axis=0)
                self._update_means(x_label, label)
        else:
            self._update_means(x, labels)

    def _stop_means(self):
        """Calculate the class means."""
        for label in self.means:
            self.means[label] /= self.tlens[label]

    def _update_means(self, x, label):
        """Update the internal variables that store the data for the means.

        x -- Data points from a single class.
        label -- The label for that class.
        """
        if label not in self.means:
            self.means[label] = numx.zeros((1, self.input_dim), dtype=self.dtype)
            self.tlens[label] = 0
        self.means[label] += x.sum(axis=0)
        self.tlens[label] += x.shape[0]

    # Training step 2: compute the overall and within-class covariance
    # matrices and solve the FDA problem

    def _train_fda(self, x, labels):
        """Gather data for the overall and within-class covariance"""
        if self._S_W is None:
            self._S_W = numx.zeros((self.input_dim, self.input_dim),
                                   dtype=self.dtype)
        # update the covariance matrix of all classes
        self._allcov.update(x)
        # if labels is a number, all x's belong to the same class
        if isinstance(labels, (list, tuple, numx.ndarray)):
            labels_ = numx.asarray(labels)
            # get all classes from labels
            for label in set(labels_):
                # group for class
                x_label = numx.compress(labels_==label, x, axis=0)
                self._update_SW(x_label, label)
        else:
            self._update_SW(x, labels)

    def _stop_fda(self):
        """Solve the eigenvalue problem for the total covariance."""
        S_T, self.avg, _ = self._allcov.fix()
        del self._allcov
        S_W = self._S_W
        del self._S_W
        # solve the generalized eigenvalue problem
        # the eigenvalues are already ordered in ascending order
        if self.output_dim is None:
            rng = None
            self.output_dim = self.input_dim
        else:
            rng = (1, self.output_dim)
        self.v = mdp.utils.symeig(S_W, S_T, range=rng, overwrite = 1)[1]

    def _update_SW(self, x, label):
        """Update the covariance matrix of the class means.

        x -- Data points from a single class.
        label -- The label for that class.
        """
        x = x - self.means[label]
        self._S_W += mdp.utils.mult(x.T, x)
    
    # Overwrite the standard methods
    
    # dummy method used to overwrite the train docstring
    def _train(self, x, label):
        """Update the internal structures according to the input data 'x'.

        x -- a matrix having different variables on different columns
            and observations on the rows.
        label -- can be a list, tuple or array of labels (one for each data
            point) or a single label, in which case all input data is assigned
            to the same class.
        """
        pass

    def _execute(self, x, n=None):
        """Compute the output of the FDA projection.

        If 'n' is an integer, then use the first 'n' components.
        """
        if n:
            v = self.v[:, 0:n]
        else:
            v = self.v
        return mdp.utils.mult(x-self.avg, v)

    def _inverse(self, y):
        return mdp.utils.mult(y, mdp.utils.pinv(self.v))+self.avg



"""
Helper script to run or debug the tests in an IDE as a simple .py file.
"""

import py

#args_str = ""
args_str = "-k hinet --maxfail 1 -s --tb native"
#args_str = "--maxfail 1 --tb native"

py.test.cmdline.main(args_str.split(" "))



"""Test the configuration object."""
from builtins import object

from mdp import config

class TestConfig(object):
    def teardown_method(self, method):
        delattr(config, 'has_test_property')

    def test_config_depfound(self):
        s = config.ExternalDepFound('test_property', 0.777)
        assert bool(s) == True
        assert config.has_test_property
        info = config.info()
        assert 'test property' in info
        assert '0.777' in info

    def test_config_depfound_string(self):
        s = config.ExternalDepFound('test_property', '0.777')
        assert bool(s) == True
        assert config.has_test_property
        info = config.info()
        assert 'test property' in info
        assert '0.777' in info

    def test_config_depfailed_exc(self):
        s = config.ExternalDepFailed('test_property', ImportError('GOOGOO'))
        assert bool(s) == False
        assert not config.has_test_property
        info = config.info()
        assert 'test property' in info
        assert 'GOOGOO' in info

    def test_config_depfailed_string(self):
        s = config.ExternalDepFailed('test_property', 'GOOGOO')
        assert bool(s) == False
        assert not config.has_test_property
        info = config.info()
        assert 'test property' in info
        assert 'GOOGOO' in info



"""
Module for parallel flows that can handle the parallel training / execution.

Corresponding classes for task callables and ResultContainer are defined here
as well.
"""
from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import next
from builtins import range

import mdp
from mdp import numx as n

from .parallelnodes import NotForkableParallelException
from .scheduling import (
    TaskCallable, ResultContainer, OrderedResultContainer, Scheduler
)
from mdp.hinet import FlowNode


### Helper code for node purging before transport. ###

class _DummyNode(mdp.Node):
    """Dummy node class for empty nodes."""

    @staticmethod
    def is_trainable():
        return False
    
    def _execute(self, x):
        err = "This is only a dummy created by 'parallel._purge_flownode'."
        raise mdp.NodeException(err)


_DUMMY_NODE = _DummyNode()

def _purge_flownode(flownode):
    """Replace nodes that are """
    for i_node, node in enumerate(flownode._flow):
        if not (node._train_phase_started or node.use_execute_fork()):
            flownode._flow.flow[i_node] = _DUMMY_NODE


### Train task classes ###

class FlowTaskCallable(TaskCallable):
    """Base class for all flow callables.

    It deals activating the required extensions.
    """

    def __init__(self):
        """Store the currently active extensions."""
        self._used_extensions = mdp.get_active_extensions()
        super(FlowTaskCallable, self).__init__()

    def setup_environment(self):
        """Activate the used extensions."""
        # deactivate all active extensions for safety
        mdp.deactivate_extensions(mdp.get_active_extensions())
        mdp.activate_extensions(self._used_extensions)


class FlowTrainCallable(FlowTaskCallable):
    """Implements a single training phase in a flow for a data block.

    A FlowNode is used to simplify the forking process and to
    encapsulate the flow.

    You can also derive from this class to define your own callable class.
    """

    def __init__(self, flownode, purge_nodes=True):
        """Store everything for the training.

        keyword arguments:
        flownode -- FlowNode containing the flow to be trained.
        purge_nodes -- If True nodes not needed for the join will be replaced
            with dummy nodes to reduce the footprint.
        """
        self._flownode = flownode
        self._purge_nodes = purge_nodes
        super(FlowTrainCallable, self).__init__()

    def __call__(self, data):
        """Do the training and return only the trained node.

        data -- training data block (array or list if additional arguments are
            required)
        """
        if type(data) is n.ndarray:
            self._flownode.train(data)
        else:
            self._flownode.train(*data)
        # note the local training in ParallelFlow relies on the flownode
        # being preserved, so derived classes should preserve it as well
        if self._purge_nodes:
            _purge_flownode(self._flownode)
        return self._flownode

    def fork(self):
        return self.__class__(self._flownode.fork(),
                              purge_nodes=self._purge_nodes)


class TrainResultContainer(ResultContainer):
    """Container for parallel nodes.

    Expects flownodes as results and joins them to save memory.
    A list containing one flownode is returned, so this container can replace
    the standard list container without any changes elsewhere.
    """

    def __init__(self):
        super(TrainResultContainer, self).__init__()
        self._flownode = None

    def add_result(self, result, task_index):
        if not self._flownode:
            self._flownode = result
        else:
            self._flownode.join(result)

    def get_results(self):
        flownode = self._flownode
        self._flownode = None
        return [flownode,]
    

### Execute task classes ###

class FlowExecuteCallable(FlowTaskCallable):
    """Implements data execution through a Flow.
    
    A FlowNode is used to simplify the forking process and to
    encapsulate the flow.
    """

    def __init__(self, flownode, nodenr=None, purge_nodes=True):
        """Store everything for the execution.

        flownode -- FlowNode for the execution
        nodenr -- optional nodenr argument for the flow execute method
        purge_nodes -- If True nodes not needed for the join will be replaced
            with dummy nodes to reduce the footprint.
        """
        self._flownode = flownode
        self._nodenr = nodenr
        self._purge_nodes = purge_nodes
        super(FlowExecuteCallable, self).__init__()

    def __call__(self, x):
        """Return the execution result.

        x -- data chunk
        
        If use_fork_execute is True for the flownode then it is returned
        in the result tuple.
        """
        y = self._flownode.execute(x, nodenr=self._nodenr)
        if self._flownode.use_execute_fork():
            if self._purge_nodes:
                _purge_flownode(self._flownode)
            return (y, self._flownode)
        else:
            return (y, None)

    def fork(self):
        return self.__class__(self._flownode.fork(), nodenr=self._nodenr,
                              purge_nodes=self._purge_nodes)
    

class ExecuteResultContainer(OrderedResultContainer):
    """Default result container with automatic restoring of the result order.

    This result container should be used together with BiFlowExecuteCallable.
    Both the execute result (x and possibly msg) and the forked BiFlowNode
    are stored.
    """

    def __init__(self):
        """Initialize attributes."""
        super(ExecuteResultContainer, self).__init__()
        self._flownode = None

    def add_result(self, result, task_index):
        """Remove the forked BiFlowNode from the result and join it."""
        excecute_result, forked_flownode = result
        super(ExecuteResultContainer, self).add_result(excecute_result,
                                                       task_index)
        if forked_flownode is not None:
            if self._flownode is None:
                self._flownode = forked_flownode
            else:
                self._flownode.join(forked_flownode)

    def get_results(self):
        """Return the ordered results.

        The joined BiFlowNode is returned in the first result list entry,
        for the following result entries BiFlowNode is set to None.
        This reduces memory consumption while staying transparent for the
        ParallelBiFlow.
        """
        excecute_results = super(ExecuteResultContainer, self).get_results()
        flownode_results = ([self._flownode,]
                              + ([None] * (len(excecute_results)-1)))
        return list(zip(excecute_results, flownode_results))

### ParallelFlow Class ###

class ParallelFlowException(mdp.FlowException):
    """Standard exception for problems with ParallelFlow."""
    pass


class NoTaskException(ParallelFlowException):
    """Exception for problems with the task creation."""
    pass


class ParallelFlow(mdp.Flow):
    """A parallel flow provides the methods for parallel training / execution.

    Nodes in the flow which are not derived from ParallelNode are trained in
    the normal way. The training is also done normally if fork() raises a
    TrainingPhaseNotParallelException. This can be intentionally used by the
    node to request local training without forking.
    Parallel execution on the other hand should work for all nodes, since it
    only relies on the copy method of nodes.
    The stop_training method is always called locally, with no forking or
    copying involved.

    Both parallel training and execution can be done conveniently by providing
    a scheduler instance to the train or execute method.
    It is also possible to manage the tasks manually. This is done via the
    methods setup_parallel_training (or execution), get_task and use_results.
    The code of the train / execute method can serve as an example how to use
    these methods and process the tasks by a scheduler.
    """

    def __init__(self, flow, verbose=False, **kwargs):
        """Initialize the internal variables.

        Note that the crash_recovery flag is is not supported, so it is
        disabled.
        """
        kwargs["crash_recovery"] = False
        super(ParallelFlow, self).__init__(flow, verbose=verbose,
                                           **kwargs)
        self._train_data_iterables = None  # all training data
        self._train_data_iterator = None  # iterator for current training
        # index of currently trained node, also used as flag for training
        # takes value None for not training
        self._i_train_node = None
        self._flownode = None  # used during training
        # iterable for execution data
        # also signals if parallel execution is underway
        self._exec_data_iterator = None
        self._next_task = None  # buffer for next task
        self._train_callable_class = None
        self._execute_callable_class = None

    @mdp.with_extension("parallel")
    def train(self, data_iterables, scheduler=None,
              train_callable_class=None,
              overwrite_result_container=True,
              **kwargs):
        """Train all trainable nodes in the flow.

        If a scheduler is provided the training will be done in parallel on the
        scheduler.

        data_iterables -- A list of iterables, one for each node in the flow.
            The iterators returned by the iterables must
            return data arrays that are then used for the node training.
            See Flow.train for more details.
            If a custom train_callable_class is used to preprocess the data
            then other data types can be used as well.
        scheduler -- Value can be either None for normal training (default
            value) or a Scheduler instance for parallel training with the
            scheduler.
            If the scheduler value is an iterable or iterator then it is
            assumed that it contains a scheduler for each training phase.
            After a node has been trained the scheduler is shutdown. Note that
            you can e.g. use a generator to create the schedulers just in time.
            For nodes which are not trained the scheduler can be None.
        train_callable_class -- Class used to create training callables for the
            scheduler. By specifying your own class you can implement data
            transformations before the data is actually fed into the flow
            (e.g. from 8 bit image to 64 bit double precision).
            Note that the train_callable_class is only used if a scheduler was
            provided. By default NodeResultContainer is used.
        overwrite_result_container -- If set to True (default value) then
            the result container in the scheduler will be overwritten with an
            instance of NodeResultContainer (unless it is already an instance
            of NodeResultContainer). This improves the memory efficiency.
        """
        # Warning: If this method is updated you also have to update train
        #          in ParallelCheckpointFlow.
        if self.is_parallel_training:
            raise ParallelFlowException("Parallel training is underway.")
        if scheduler is None:
            if train_callable_class is not None:
                err = ("A train_callable_class was specified but no scheduler "
                       "was given, so the train_callable_class has no effect.")
                raise ParallelFlowException(err)
            super(ParallelFlow, self).train(data_iterables, **kwargs)
        else:
            if train_callable_class is None:
                train_callable_class = FlowTrainCallable
            schedulers = None
            # do parallel training
            try:
                self.setup_parallel_training(
                                    data_iterables,
                                    train_callable_class=train_callable_class,
                                    **kwargs)
                # prepare scheduler
                if not isinstance(scheduler, Scheduler):
                    # scheduler contains an iterable with the schedulers
                    # self._i_train_node was set in setup_parallel_training
                    schedulers = iter(scheduler)
                    scheduler = next(schedulers)
                    if self._i_train_node > 0:
                        # dispose schedulers for pretrained nodes
                        for _ in range(self._i_train_node):
                            if scheduler is not None:
                                scheduler.shutdown()
                            scheduler = next(schedulers)
                    elif self._i_train_node is None:
                        # all nodes are already trained, dispose schedulers
                        for _ in range(len(self.flow) - 1):
                            if scheduler is not None:
                                scheduler.shutdown()
                            # the last scheduler will be shutdown in finally
                            scheduler = next(schedulers)
                    last_trained_node = self._i_train_node
                else:
                    schedulers = None
                # check that the scheduler is compatible
                if ((scheduler is not None) and
                    overwrite_result_container and
                    (not isinstance(scheduler.result_container,
                                    TrainResultContainer))):
                    scheduler.result_container = TrainResultContainer()
                ## train all nodes
                while self.is_parallel_training:
                    while self.task_available:
                        task = self.get_task()
                        scheduler.add_task(*task)
                    results = scheduler.get_results()
                    if results == []:
                        err = ("Could not get any training tasks or results "
                               "for the current training phase.")
                        raise Exception(err)
                    else:
                        self.use_results(results)
                    # check if we have to switch to next scheduler
                    if ((schedulers is not None) and
                        (self._i_train_node is not None) and
                        (self._i_train_node > last_trained_node)):
                        # dispose unused schedulers
                        for _ in range(self._i_train_node - last_trained_node):
                            if scheduler is not None:
                                scheduler.shutdown()
                            scheduler = next(schedulers)
                        last_trained_node = self._i_train_node
                        # check that the scheduler is compatible
                        if ((scheduler is not None) and
                            overwrite_result_container and
                            (not isinstance(scheduler.result_container,
                                            TrainResultContainer))):
                            scheduler.result_container = TrainResultContainer()
            finally:
                # reset iterable references, which cannot be pickled
                self._train_data_iterables = None
                self._train_data_iterator = None
                if (schedulers is not None) and (scheduler is not None):
                    scheduler.shutdown()

    def setup_parallel_training(self, data_iterables,
                                train_callable_class=FlowTrainCallable):
        """Prepare the flow for handing out tasks to do the training.

        After calling setup_parallel_training one has to pick up the
        tasks with get_task, run them and finally return the results via
        use_results. tasks are available as long as task_available returns
        True. Training may require multiple phases, which are each closed by
        calling use_results.

        data_iterables -- A list of iterables, one for each node in the flow.
            The iterators returned by the iterables must
            return data arrays that are then used for the node training.
            See Flow.train for more details.
            If a custom train_callable_class is used to preprocess the data
            then other data types can be used as well.
        train_callable_class -- Class used to create training callables for the
            scheduler. By specifying your own class you can implement data
            transformations before the data is actually fed into the flow
            (e.g. from 8 bit image to 64 bit double precision).
        """
        if self.is_parallel_training:
            err = "Parallel training is already underway."
            raise ParallelFlowException(err)
        self._train_callable_class = train_callable_class
        self._train_data_iterables = self._train_check_iterables(data_iterables)
        self._i_train_node = 0
        self._flownode = FlowNode(mdp.Flow(self.flow))
        self._next_train_phase()

    def _next_train_phase(self):
        """Find the next phase or node for parallel training.

        When it is found the corresponding internal variables are set.
        Nodes which are not derived from ParallelNode are trained locally.
        If a fork() fails due to a TrainingPhaseNotParallelException
        in a certain train phase, then the training is done locally as well
        (but fork() is tested again for the next phase).
        """
        # find next node that can be forked, if required do local training
        while self._i_train_node < len(self.flow):
            current_node = self.flow[self._i_train_node]
            if not current_node.is_training():
                self._i_train_node += 1
                continue
            data_iterable = self._train_data_iterables[self._i_train_node]
            try:
                self._flownode.fork()
                # fork successful, prepare parallel training
                if self.verbose:
                    print ("start parallel training phase of " +
                           "node no. %d in parallel flow" %
                           (self._i_train_node+1))
                self._train_data_iterator = iter(data_iterable)
                first_task = self._create_train_task()
                # make sure that the iterator is not empty
                if first_task is None:
                    if current_node.get_current_train_phase() == 1:
                        err_str = ("The training data iteration for node "
                                   "no. %d could not be repeated for the "
                                   "second training phase, you probably "
                                   "provided an iterator instead of an "
                                   "iterable." % (self._i_train_node+1))
                        raise mdp.FlowException(err_str)
                    else:
                        err_str = ("The training data iterator for node "
                                   "no. %d is empty." % (self._i_train_node+1))
                        raise mdp.FlowException(err_str)
                task_data_chunk = first_task[0]
                # Only first task contains the new callable (enable caching).
                # A fork is not required here, since the callable is always
                # forked in the scheduler.
                self._next_task = (task_data_chunk,
                                   self._train_callable_class(self._flownode))
                break
            except NotForkableParallelException as exception:
                if self.verbose:
                    print ("could not fork node no. %d: %s" %
                           (self._i_train_node+1, str(exception)))
                    print ("start nonparallel training phase of " +
                           "node no. %d in parallel flow" %
                           (self._i_train_node+1))
                self._local_train_phase(data_iterable)
                if self.verbose:
                    print ("finished nonparallel training phase of " +
                           "node no. %d in parallel flow" %
                           (self._i_train_node+1))
                self._stop_training_hook()
                self._flownode.stop_training()
                self._post_stop_training_hook()
                if not self.flow[self._i_train_node].is_training():
                    self._i_train_node += 1
        else:
            # training is finished
            self._i_train_node = None

    def _local_train_phase(self, data_iterable):
        """Perform a single training phase locally.

        The internal _train_callable_class is used for the training.
        """
        current_node = self.flow[self._i_train_node]
        task_callable = self._train_callable_class(self._flownode,
                                                   purge_nodes=False)
        empty_iterator = True
        for i_task, data in enumerate(data_iterable):
            empty_iterator = False
            # Note: if x contains additional args assume that the
            # callable can handle this
            task_callable(data)
            if self.verbose:
                print ("    finished nonparallel task no. %d" % (i_task+1))
        if empty_iterator:
            if current_node.get_current_train_phase() == 1:
                err_str = ("The training data iteration for node "
                           "no. %d could not be repeated for the "
                           "second training phase, you probably "
                           "provided an iterator instead of an "
                           "iterable." % (self._i_train_node+1))
                raise mdp.FlowException(err_str)
            else:
                err_str = ("The training data iterator for node "
                           "no. %d is empty." % (self._i_train_node+1))
                raise mdp.FlowException(err_str)

    def _post_stop_training_hook(self):
        """Hook method that is called after stop_training is called."""
        pass

    def _create_train_task(self):
        """Create and return a single training task without callable.

        Returns None if data iterator end is reached.
        """
        try:
            return (next(self._train_data_iterator), None)
        except StopIteration:
            return None

    @mdp.with_extension("parallel")
    def execute(self, iterable, nodenr=None, scheduler=None,
                execute_callable_class=None,
                overwrite_result_container=True):
        """Train all trainable nodes in the flow.

        If a scheduler is provided the execution will be done in parallel on
        the scheduler.

        iterable -- An iterable or iterator that returns data arrays that are
            used as input to the flow. Alternatively, one can specify one
            data array as input.
            If a custom execute_callable_class is used to preprocess the data
            then other data types can be used as well.
        nodenr -- Same as in normal flow, the flow is only executed up to the
            nodenr.
        scheduler -- Value can be either None for normal execution (default
            value) or a Scheduler instance for parallel execution with the
            scheduler.
        execute_callable_class -- Class used to create execution callables for
            the scheduler. By specifying your own class you can implement data
            transformations before the data is actually fed into the flow
            (e.g. from 8 bit image to 64 bit double precision).
            Note that the execute_callable_class is only used if a scheduler was
            provided. If a scheduler is provided the default class used is
            NodeResultContainer.
        overwrite_result_container -- If set to True (default value) then
            the result container in the scheduler will be overwritten with an
            instance of OrderedResultContainer (unless it is already an
            instance of OrderedResultContainer). Otherwise the results might
            have a different order than the data chunks, which could mess up
            any subsequent analysis.
        """
        if self.is_parallel_training:
            raise ParallelFlowException("Parallel training is underway.")
        if scheduler is None:
            if execute_callable_class is not None:
                err = ("A execute_callable_class was specified but no "
                       "scheduler was given, so the execute_callable_class "
                       "has no effect.")
                raise ParallelFlowException(err)
            return super(ParallelFlow, self).execute(iterable, nodenr)
        if execute_callable_class is None:
            execute_callable_class = FlowExecuteCallable
        # check that the scheduler is compatible
        if overwrite_result_container:
            if not isinstance(scheduler.result_container,
                              ExecuteResultContainer):
                scheduler.result_container = ExecuteResultContainer()
        # do parallel execution
        self._flownode = FlowNode(mdp.Flow(self.flow))
        try:
            self.setup_parallel_execution(
                                iterable,
                                nodenr=nodenr,
                                execute_callable_class=execute_callable_class)
            while self.task_available:
                task = self.get_task()
                scheduler.add_task(*task)
            result = self.use_results(scheduler.get_results())
        finally:
            # reset remaining iterator references, which cannot be pickled
            self._exec_data_iterator = None
        return result

    def setup_parallel_execution(self, iterable, nodenr=None,
                                 execute_callable_class=FlowExecuteCallable):
        """Prepare the flow for handing out tasks to do the execution.

        After calling setup_parallel_execution one has to pick up the
        tasks with get_task, run them and finally return the results via
        use_results. use_results will then return the result as if the flow was
        executed in the normal way.

        iterable -- An iterable or iterator that returns data arrays that are
            used as input to the flow. Alternatively, one can specify one
            data array as input.
            If a custom execute_callable_class is used to preprocess the data
            then other data types can be used as well.
        nodenr -- Same as in normal flow, the flow is only executed up to the
            nodenr.
        execute_callable_class -- Class used to create execution callables for
            the scheduler. By specifying your own class you can implement data
            transformations before the data is actually fed into the flow
            (e.g. from 8 bit image to 64 bit double precision).
        """
        if self.is_parallel_training:
            raise ParallelFlowException("Parallel training is underway.")
        self._execute_callable_class = execute_callable_class
        if isinstance(iterable, n.ndarray):
            iterable = [iterable]
        self._exec_data_iterator = iter(iterable)
        first_task = self._create_execute_task()
        if first_task is None:
            errstr = ("The execute data iterator is empty.")
            raise mdp.FlowException(errstr)
        task_data_chunk = first_task[0]
        # Only first task contains the new callable (enable caching).
        # A fork is not required here, since the callable is always
        # forked in the scheduler.
        self._next_task = (task_data_chunk,
                           self._execute_callable_class(self._flownode,
                                                        purge_nodes=True))

    def _create_execute_task(self):
        """Create and return a single execution task.

        Returns None if data iterator end is reached.
        """
        try:
            # TODO: check if forked task is forkable before enforcing caching
            return (next(self._exec_data_iterator), None)
        except StopIteration:
            return None

    def get_task(self):
        """Return a task either for either training or execution.

        A a one task buffer is used to make task_available work.
        tasks are available as long as need_result returns False or all the
        training / execution is done. If no tasks are available a
        NoTaskException is raised.
        """
        if self._next_task is not None:
            task = self._next_task
            if self._i_train_node is not None:
                self._next_task = self._create_train_task()
            elif self._exec_data_iterator is not None:
                self._next_task = self._create_execute_task()
            else:
                raise NoTaskException("No data available for execution task.")
            return task
        else:
            raise NoTaskException("No task available for execution.")

    @property
    def is_parallel_training(self):
        """Return True if parallel training is underway."""
        return self._i_train_node is not None

    @property
    def is_parallel_executing(self):
        """Return True if parallel execution is underway."""
        return self._exec_data_iterator is not None

    @property
    def task_available(self):
        """Return True if tasks are available, otherwise False.

        If False is returned this can indicate that results are needed to
        continue training.
        """
        return self._next_task is not None

    def use_results(self, results):
        """Use the result from the scheduler.

        During parallel training this will start the next training phase.
        For parallel execution this will return the result, like a normal
        execute would.

        results -- Iterable containing the results, normally the return value
            of scheduler.ResultContainer.get_results().
            The individual results can be the return values of the tasks.
        """
        if self.is_parallel_training:
            for result in results:
                # the flownode contains the original nodes
                self._flownode.join(result)
            if self.verbose:
                print ("finished parallel training phase of node no. " +
                       "%d in parallel flow" % (self._i_train_node+1))
            self._stop_training_hook()
            self._flownode.stop_training()
            self._post_stop_training_hook()
            if not self.flow[self._i_train_node].is_training():
                self._i_train_node += 1
            self._next_train_phase()
        elif self.is_parallel_executing:
            self._exec_data_iterator = None
            ys = [result[0] for result in results]
            if self._flownode.use_execute_fork():
                flownodes = [result[1] for result in results]
                for flownode in flownodes:
                    if flownode is not None:
                        self._flownode.join(flownode)
            return n.concatenate(ys)


class ParallelCheckpointFlow(ParallelFlow, mdp.CheckpointFlow):
    """Parallel version of CheckpointFlow.

    Note that train phases are always closed, so e.g. CheckpointSaveFunction
    should not expect open train phases. This is necessary since otherwise
    stop_training() would be called remotely.
    """

    def __init__(self, flow, verbose=False, **kwargs):
        """Initialize the internal variables."""
        self._checkpoints = None
        super(ParallelCheckpointFlow, self).__init__(flow=flow,
                                                     verbose=verbose,
                                                     **kwargs)

    def train(self, data_iterables, checkpoints, scheduler=None,
              train_callable_class=FlowTrainCallable,
              overwrite_result_container=True,
              **kwargs):
        """Train all trainable nodes in the flow.

        Same as the train method in ParallelFlow, but with additional support
        of checkpoint functions as in CheckpointFlow.
        """
        super(ParallelCheckpointFlow, self).train(
                        data_iterables=data_iterables,
                        scheduler=scheduler,
                        train_callable_class=train_callable_class,
                        overwrite_result_container=overwrite_result_container,
                        checkpoints=checkpoints,
                        **kwargs)

    def setup_parallel_training(self, data_iterables, checkpoints,
                                train_callable_class=FlowTrainCallable,
                                **kwargs):
        """Checkpoint version of parallel training."""
        self._checkpoints = self._train_check_checkpoints(checkpoints)
        super(ParallelCheckpointFlow, self).setup_parallel_training(
                                    data_iterables,
                                    train_callable_class=train_callable_class,
                                    **kwargs)

    def _post_stop_training_hook(self):
        """Check if we reached a checkpoint."""
        super(ParallelCheckpointFlow, self)._post_stop_training_hook()
        i_node = self._i_train_node
        if self.flow[i_node].get_remaining_train_phase() == 0:
            if ((i_node <= len(self._checkpoints))
                and self._checkpoints[i_node]):
                dict = self._checkpoints[i_node](self.flow[i_node])
                # store result, just like in the original CheckpointFlow
                if dict:
                    self.__dict__.update(dict)



"""
Module for Layers.

Note that additional args and kwargs for train or execute are currently not
supported.
"""

import mdp
from mdp import numx

# TODO: maybe turn self.nodes into a read only property with self._nodes

# TODO: Find a better way to deal with additional args for train/execute?
#    Maybe split them by default, but can be disabled via switch?

class Layer(mdp.Node):
    """Layers are nodes which consist of multiple horizontally parallel nodes.

    The incoming data is split up according to the dimensions of the internal
    nodes. For example if the first node has an input_dim of 50 and the second
    node 100 then the layer will have an input_dim of 150. The first node gets
    x[:,:50], the second one x[:,50:].

    Any additional arguments are forwarded unaltered to each node.
    Warning: This might change in the next release (2.5).

    Since they are nodes themselves layers can be stacked in a flow (e.g. to
    build a layered network). If one would like to use flows instead of nodes
    inside of a layer one can use a FlowNode.
    """

    def __init__(self, nodes, dtype=None):
        """Setup the layer with the given list of nodes.

        The input and output dimensions for the nodes must be already set
        (the output dimensions for simplicity reasons). The training phases for
        the nodes are allowed to differ.

        Keyword arguments:
        nodes -- List of the nodes to be used.
        """
        self.nodes = nodes
        # check nodes properties and get the dtype
        dtype = self._check_props(dtype)
        # calculate the the dimensions
        self.node_input_dims = numx.zeros(len(self.nodes))
        input_dim = 0
        for index, node in enumerate(nodes):
            input_dim += node.input_dim
            self.node_input_dims[index] = node.input_dim
        output_dim = self._get_output_dim_from_nodes()
        super(Layer, self).__init__(input_dim=input_dim,
                                    output_dim=output_dim,
                                    dtype=dtype)

    def _get_output_dim_from_nodes(self):
        """Calculate the output_dim from the nodes and return it.

        If the output_dim of a node is not set the None is returned.
        """
        output_dim = 0
        for node in self.nodes:
            if node.output_dim is not None:
                output_dim += node.output_dim
            else:
                return None
        return output_dim

    def _check_props(self, dtype):
        """Check the compatibility of the properties of the internal nodes.

        Return the found dtype and check the dimensions.

        dtype -- The specified layer dtype.
        """
        dtype_list = []  # the dtypes for all the nodes
        for i, node in enumerate(self.nodes):
            # input_dim for each node must be set
            if node.input_dim is None:
                err = ("input_dim must be set for every node. " +
                       "Node #%d (%s) does not comply." % (i, node))
                raise mdp.NodeException(err)
            if node.dtype is not None:
                dtype_list.append(node.dtype)
        # check that the dtype is None or the same for every node
        nodes_dtype = None
        nodes_dtypes = set(dtype_list)
        nodes_dtypes.discard(None)
        if len(nodes_dtypes) > 1:
            err = ("All nodes must have the same dtype (found: %s)." %
                   nodes_dtypes)
            raise mdp.NodeException(err)
        elif len(nodes_dtypes) == 1:
            nodes_dtype = list(nodes_dtypes)[0]
        # check that the nodes dtype matches the specified dtype
        if nodes_dtype and dtype:
            if not numx.dtype(nodes_dtype) == numx.dtype(dtype):
                err = ("Cannot set dtype to %s: " %
                       numx.dtype(nodes_dtype).name +
                       "an internal node requires %s" % numx.dtype(dtype).name)
                raise mdp.NodeException(err)
        elif nodes_dtype and not dtype:
            dtype = nodes_dtype
        return dtype

    def _set_dtype(self, t):
        for node in self.nodes:
            node.dtype = t
        self._dtype = t

    def _get_supported_dtypes(self):
        # we supported the minimal common dtype set
        types = set(mdp.utils.get_dtypes('All'))
        for node in self.nodes:
            types = types.intersection(node.get_supported_dtypes())
        return list(types)

    def is_trainable(self):
        return any(node.is_trainable() for node in self.nodes)

    def is_invertible(self):
        return all(node.is_invertible() for node in self.nodes)

    def _get_train_seq(self):
        """Return the train sequence.

        The length is set by the node with maximum length.
        """
        max_train_length = 0
        for node in self.nodes:
            node_length = len(node._get_train_seq())
            if node_length > max_train_length:
                max_train_length = node_length
        return ([[self._train, self._stop_training]] * max_train_length)

    def _train(self, x, *args, **kwargs):
        """Perform single training step by training the internal nodes."""
        start_index = 0
        stop_index = 0
        for node in self.nodes:
            start_index = stop_index
            stop_index += node.input_dim
            if node.is_training():
                node.train(x[:, start_index : stop_index], *args, **kwargs)

    def _stop_training(self, *args, **kwargs):
        """Stop training of the internal nodes."""
        for node in self.nodes:
            if node.is_training():
                node.stop_training(*args, **kwargs)
        if self.output_dim is None:
            self.output_dim = self._get_output_dim_from_nodes()

    def _pre_execution_checks(self, x):
        """Make sure that output_dim is set and then perform normal checks."""
        if self.output_dim is None:
            # first make sure that the output_dim is set for all nodes
            in_start = 0
            in_stop = 0
            for node in self.nodes:
                in_start = in_stop
                in_stop += node.input_dim
                node._pre_execution_checks(x[:,in_start:in_stop])
            self.output_dim = self._get_output_dim_from_nodes()
            if self.output_dim is None:
                err = "output_dim must be set at this point for all nodes"
                raise mdp.NodeException(err)
        super(Layer, self)._pre_execution_checks(x)

    def _execute(self, x, *args, **kwargs):
        """Process the data through the internal nodes."""
        in_start = 0
        in_stop = 0
        out_start = 0
        out_stop = 0
        y = None
        for node in self.nodes:
            out_start = out_stop
            out_stop += node.output_dim
            in_start = in_stop
            in_stop += node.input_dim
            if y is None:
                node_y = node.execute(x[:,in_start:in_stop], *args, **kwargs)
                y = numx.zeros([node_y.shape[0], self.output_dim],
                               dtype=node_y.dtype)
                y[:,out_start:out_stop] = node_y
            else:
                y[:,out_start:out_stop] = node.execute(x[:,in_start:in_stop],
                                                        *args, **kwargs)
        return y

    def _inverse(self, x, *args, **kwargs):
        """Combine the inverse of all the internal nodes."""
        in_start = 0
        in_stop = 0
        out_start = 0
        out_stop = 0
        y = None
        for node in self.nodes:
            # compared with execute, input and output are switched
            out_start = out_stop
            out_stop += node.input_dim
            in_start = in_stop
            in_stop += node.output_dim
            if y is None:
                node_y = node.inverse(x[:,in_start:in_stop], *args, **kwargs)
                y = numx.zeros([node_y.shape[0], self.input_dim],
                               dtype=node_y.dtype)
                y[:,out_start:out_stop] = node_y
            else:
                y[:,out_start:out_stop] = node.inverse(x[:,in_start:in_stop],
                                                        *args, **kwargs)
        return y

    ## container methods ##

    def __len__(self):
        return len(self.nodes)

    def __getitem__(self, key):
        return self.nodes.__getitem__(key)

    def __contains__(self, item):
        return self.nodes.__contains__(item)

    def __iter__(self):
        return self.nodes.__iter__()


class CloneLayer(Layer):
    """Layer with a single node instance that is used multiple times.

    The same single node instance is used to build the layer, so
    Clonelayer(node, 3) executes in the same way as Layer([node]*3).
    But Layer([node]*3) would have a problem when closing a training phase,
    so one has to use CloneLayer.

    A CloneLayer can be used for weight sharing in the training phase. It might
    be also useful for reducing the memory footprint use during the execution
    phase (since only a single node instance is needed).
    """

    def __init__(self, node, n_nodes=1, dtype=None):
        """Setup the layer with the given list of nodes.

        Keyword arguments:
        node -- Node to be cloned.
        n_nodes -- Number of repetitions/clones of the given node.
        """
        super(CloneLayer, self).__init__((node,) * n_nodes, dtype=dtype)
        self.node = node  # attribute for convenience

    def _stop_training(self, *args, **kwargs):
        """Stop training of the internal node."""
        if self.node.is_training():
            self.node.stop_training(*args, **kwargs)
        if self.output_dim is None:
            self.output_dim = self._get_output_dim_from_nodes()

class SameInputLayer(Layer):
    """SameInputLayer is a layer were all nodes receive the full input.

    So instead of splitting the input according to node dimensions, all nodes
    receive the complete input data.
    """

    def __init__(self, nodes, dtype=None):
        """Setup the layer with the given list of nodes.

        The input dimensions for the nodes must all be equal, the output
        dimensions can differ (but must be set as well for simplicity reasons).

        Keyword arguments:
        nodes -- List of the nodes to be used.
        """
        self.nodes = nodes
        # check node properties and get the dtype
        dtype = self._check_props(dtype)
        # check that the input dimensions are all the same
        input_dim = self.nodes[0].input_dim
        for node in self.nodes:
            if not node.input_dim == input_dim:
                err = "The nodes have different input dimensions."
                raise mdp.NodeException(err)
        output_dim = self._get_output_dim_from_nodes()
        # intentionally use MRO above Layer, not SameInputLayer
        super(Layer, self).__init__(input_dim=input_dim,
                                    output_dim=output_dim,
                                    dtype=dtype)

    @staticmethod
    def is_invertible():
        return False

    def _train(self, x, *args, **kwargs):
        """Perform single training step by training the internal nodes."""
        for node in self.nodes:
            if node.is_training():
                node.train(x, *args, **kwargs)

    def _pre_execution_checks(self, x):
        """Make sure that output_dim is set and then perform nromal checks."""
        if self.output_dim is None:
            # first make sure that the output_dim is set for all nodes
            for node in self.nodes:
                node._pre_execution_checks(x)
            self.output_dim = self._get_output_dim_from_nodes()
            if self.output_dim is None:
                err = "output_dim must be set at this point for all nodes"
                raise mdp.NodeException(err)
        # intentionally use MRO above Layer, not SameInputLayer
        super(Layer, self)._pre_execution_checks(x)

    def _execute(self, x, *args, **kwargs):
        """Process the data through the internal nodes."""
        out_start = 0
        out_stop = 0
        y = None
        for node in self.nodes:
            out_start = out_stop
            out_stop += node.output_dim
            if y is None:
                node_y = node.execute(x, *args, **kwargs)
                y = numx.zeros([node_y.shape[0], self.output_dim],
                               dtype=node_y.dtype)
                y[:,out_start:out_stop] = node_y
            else:
                y[:,out_start:out_stop] = node.execute(x, *args, **kwargs)
        return y



"""
Module to automatically create a module with BiMDP versions of MDP nodes.
"""
from future import standard_library
standard_library.install_aliases()
from builtins import str

import inspect
import mdp
from io import StringIO

# Blacklist of nodes that cause problems with autogeneration
NOAUTOGEN_MDP_NODES = [
    "NoiseNode"  # function default value causes trouble
]
NOAUTOGEN_MDP_CLASSIFIERS = []


def _get_node_subclasses(node_class=mdp.Node, module=mdp.nodes):
    """
    Return all node classes in module which are subclasses of node_class.
    """
    node_subclasses = []
    for node_subclass in (getattr(module, name) for name in dir(module)):
        if (isinstance(node_subclass, type) and
            issubclass(node_subclass, node_class)):
            node_subclasses.append(node_subclass)
    return node_subclasses


def _binode_code(write, node_class, modulename, base_classname="BiNode",
                 old_classname="Node"):
    """Write code for BiMDP versions of normal node classes into module file.
    
    It preserves the signature, which is useful for introspection (this is
    used by the ParallelNode _default_fork implementation).

    write -- Function to write out code.
    node_class -- Node class for which the new node class will be created.
    modulename -- Name of the module where the node is from.
    base_classname -- Base class to be used for the new nodes.
    old_classname -- Name of the original base class, which will be replaced
        in the new class name.
    """
    node_name = node_class.__name__
    binode_name = node_name[:-len(old_classname)] + base_classname
    write('class %s(%s, %s.%s):' %
          (binode_name, base_classname, modulename, node_name))
    docstring = ("Automatically created %s version of %s." %
                 (base_classname, node_name))
    write('\n    """%s"""' % docstring)
    ## define the init method explicitly to preserve the signature
    docstring = node_class.__init__.__doc__
    args, varargs, varkw, defaults = inspect.getargspec(node_class.__init__)
    args.remove('self')
    args += ('node_id', 'stop_result')
    defaults += (None, None)
    if defaults is None:
        defaults = []
    first_default = len(args) - len(defaults)
    write('\n    def __init__(self')
    write(''.join(', ' + arg for arg in args[:-len(defaults)]))
    write(''.join(', ' + arg + '=' + repr(defaults[i_arg])
                  for i_arg, arg in enumerate(args[first_default:])))
    if varargs:
        write(', *%s' % varargs)
    # always support kwargs, to prevent multiple-inheritance issues
    if not varkw:
        varkw = "kwargs"
    write(', **%s' % varkw)
    write('):')
    if docstring:
        write('\n        """%s"""' % docstring)
    write('\n        super(%s, self).__init__(' % binode_name)
    write(', '.join('%s=%s' % (arg, arg) for arg in args))
    if varargs:
        if args:
            write(', ')
        write('*%s' % varargs)
    if args or varargs:
        write(', ')
    write('**%s' % varkw)
    write(')\n\n')


def _binode_module(write, node_classes, modulename="mdp.nodes",
                   base_classname="BiNode", old_classname="Node",
                   base_import="from bimdp import BiNode"):
    """Write code for BiMDP versions of normal node classes into module file.

    write -- Function to write out code.
    node_classes -- List of node classes for which binodes are created.
    modulename -- Name of the module where the node is from.
    base_classname -- Base class to be used for the new nodes.
    old_classname -- Name of the original base class, which will be replaced
        in the new class name.
    base_import -- Inmport line for the base_class.
    """
    write('"""\nAUTOMATICALLY GENERATED CODE, DO NOT MODIFY!\n\n')
    write('Edit and run autogen.py instead to overwrite this module.\n"""')
    write('\n\nimport %s\n' % modulename)
    write(base_import + '\n\n')
    for node_class in node_classes:
        _binode_code(write, node_class, modulename,
                     base_classname=base_classname,
                     old_classname=old_classname)


def _get_unicode_write(fid):
    def write(txt):
        if type(txt) is str:
            fid.write(txt)
        else:
            fid.write(str(txt, encoding='utf-8'))
    return write


def binodes_code():
    """Generate and import the BiNode wrappers for MDP Nodes."""
    fid = StringIO()
    nodes = (node for node in
                _get_node_subclasses(node_class=mdp.Node, module=mdp.nodes)
             if not issubclass(node, mdp.ClassifierNode) and
                 node.__name__ not in NOAUTOGEN_MDP_NODES)
    _binode_module(_get_unicode_write(fid), nodes)
    return fid.getvalue()
    
def biclassifiers_code():
    """Generate and import the BiClassifier wrappers for ClassifierNodes."""
    fid = StringIO()
    def write(txt):
        if type(txt) is str:
            fid.write(txt)
        else:
            fid.write(str(txt, encoding='utf-8'))
    nodes = (node for node in
                _get_node_subclasses(node_class=mdp.ClassifierNode,
                                     module=mdp.nodes)
             if node.__name__ not in NOAUTOGEN_MDP_CLASSIFIERS)
    _binode_module(_get_unicode_write(fid), nodes,
                   base_classname="BiClassifier",
                   old_classname="Classifier",
                   base_import="from bimdp import BiClassifier")
    return fid.getvalue()



import mdp.hinet as mh
from ._tools import *
from .test_nodes_generic import (
    generic_test_factory,
    test_dtype_consistency,
    # this test fails due to the checks in _set_output_dim
    #test_outputdim_consistency,
    test_dimdtypeset,
    #test_inverse,
    # ???: test_inverse is not on the list, because it would
    # skip all the nodes in NODES anyway, because they're not
    # always invertible
)

def _get_new_flow():
    return mdp.Flow([mdp.nodes.NoiseNode(),
                     mdp.nodes.SFANode()])

def _get_new_nodes():
    return [mdp.nodes.CuBICANode(input_dim=1, whitened=True),
            mdp.nodes.CuBICANode(input_dim=2, whitened=True),
            mdp.nodes.CuBICANode(input_dim=1, whitened=True)]

def _get_single_node():
    return mdp.nodes.CuBICANode(input_dim=2, whitened=True)

def hinet_get_random_mix():
    return get_random_mix(type='d', mat_dim=(500,4))[2]

NODES = [dict(klass=mh.FlowNode,
              inp_arg_gen=hinet_get_random_mix,
              init_args = [_get_new_flow]),
         dict(klass=mh.Layer,
              inp_arg_gen=hinet_get_random_mix,
              init_args = [_get_new_nodes]),
         dict(klass=mh.CloneLayer,
              inp_arg_gen=hinet_get_random_mix,
              init_args = [_get_single_node, 2]),
         ]

def pytest_generate_tests(metafunc):
    generic_test_factory(NODES, metafunc)



from __future__ import division
from builtins import range
from past.utils import old_div
import mdp
from ._tools import *

def testRBFExpansionNode():
    rrep = mdp.utils.rrep
    dim, n = 2, 10
    centers = numx_rand.random((n, dim))
    # grid of points to numerically compute the integral
    grid = numx.meshgrid(numx.linspace(-3., 4., 100),
                         numx.linspace(-3., 4., 100))
    grid = numx.array([grid[0].flatten(), grid[1].flatten()]).T
    # compute covariance for each point of the grid
    grid_cov = numx.zeros((grid.shape[0], dim, dim))
    for i in range(dim):
        for j in range(dim):
            grid_cov[:,i,j] = grid[:,i]*grid[:,j]

    def check_mn_cov(rbf, real_covs):
        y = rbf(grid)
        # verify means, sizes
        for i in range(n):
            p = old_div(y[:,i],y[:,i].sum())
            # check mean
            mn = (rrep(p,dim)*grid).sum(0)
            assert_array_almost_equal(mn, centers[i,:], 2)
            # check variance
            vr = ((rrep(rrep(p,2),2)*grid_cov).sum(0)
                  - numx.outer(mn, mn))
            assert_array_almost_equal(vr, real_covs[i], 2)

    def scalar_to_covs(x, n):
        if numx.isscalar(x):
            x = [x]*n
        return [numx.array([[x[i],0],[0,x[i]]]) for i in range(n)]

    # 1: sizes is a scalar
    sizes = 0.32
    rbf = mdp.nodes.RBFExpansionNode(centers, sizes)
    check_mn_cov(rbf, scalar_to_covs(sizes, n))

    # 2: sizes is many scalars
    sizes = 0.3 + numx_rand.random(n)*0.2
    rbf = mdp.nodes.RBFExpansionNode(centers, sizes)
    check_mn_cov(rbf, scalar_to_covs(sizes, n))

    # 3: sizes is one covariance
    sizes = mdp.utils.symrand(numx.array([0.2, 0.4]))
    rbf = mdp.nodes.RBFExpansionNode(centers, sizes)
    check_mn_cov(rbf, [sizes]*n)

    # 4: sizes is many covariances
    sizes = [mdp.utils.symrand(numx.array([0.2, 0.4]))
             for i in range(n)]
    rbf = mdp.nodes.RBFExpansionNode(centers, sizes)
    check_mn_cov(rbf, sizes)




from builtins import str

import mdp
import mdp.hinet as hinet
n = mdp.numx

from bimdp import BiNode, BiNodeException
from bimdp import BiFlow, BiFlowException

# TODO: add derived BiFlowNode which allow specification message flag for
#    BiFlowNode to specify the internal target? Or hardwired target?


class BiFlowNode(BiNode, hinet.FlowNode):
    """BiFlowNode wraps a BiFlow of Nodes into a single BiNode.

    This is handy if you want to use a flow where a Node is required.
    Additional args and kwargs for train and execute are supported.

    Note that for nodes in the internal flow the intermediate training phases
    will generally be closed, e.g. a CheckpointSaveFunction should not expect
    these training  phases to be left open.

    All the read-only container slots are supported and are forwarded to the
    internal flow.
    """

    def __init__(self, biflow, input_dim=None, output_dim=None, dtype=None,
                 node_id=None):
        """Wrap the given BiFlow into this node.

        Pretrained nodes are allowed, but the internal _flow should not
        be modified after the BiFlowNode was created (this will cause problems
        if the training phase structure of the internal nodes changes).

        The node dimensions do not have to be specified. Unlike in a normal
        FlowNode they cannot be extracted from the nodes and are left unfixed.
        The data type is left unfixed as well.
        """
        if not isinstance(biflow, BiFlow):
            raise BiNodeException("The biflow has to be an BiFlow instance.")
        super(BiFlowNode, self).__init__(flow=biflow,
                                         input_dim=input_dim,
                                         output_dim=output_dim, dtype=dtype,
                                         node_id=node_id)
        # last successful request for target node_id
        self._last_id_request = None

    def _get_target(self):
        """Return the last successfully requested target node_id.

        The stored target is then reset to None.
        If no target is stored (i.e. if it is None) then 0 is returned.
        """
        if self._last_id_request:
            target = self._last_id_request
            self._last_id_request = None
            return target
        else:
            return 0
        return 0

    def _get_method(self, method_name, default_method, target):
        """Return the default method and the target.

        This method overrides the standard BiNode _get_method to delegate the
        method selection to the internal nodes. If the method_name is
        'inverse' then adjustments are made so that the last internal node is
        called.
        """
        if method_name == "inverse":
            if self._last_id_request is None:
                if target == -1:
                    target = None
                if target is None:
                    self._last_id_request = len(self._flow) - 1
        return default_method, target

    def _execute(self, x, msg=None):
        target = self._get_target()
        i_node = self._flow._target_to_index(target)
        # we know that _get_target returned a valid target, so no check
        return self._flow._execute_seq(x, msg, i_node)

    def _get_execute_method(self, x, method_name, target):
        """Return _execute and the provided target.

        The method selection is done in the contained nodes.
        """
        return self._execute, target

    def _get_train_seq(self):
        """Return a training sequence containing all training phases."""
        train_seq = []
        for i_node, node in enumerate(self._flow):
            if node.is_trainable():
                remaining_len = (len(node._get_train_seq())
                                 - self._pretrained_phase[i_node])
                train_seq += ([(self._get_train_function(i_node),
                                self._get_stop_training_function(i_node))]
                              * remaining_len)
        # if the last node is trainable we have to set the output dimensions
        # to those of the BiFlowNode.
        if self._flow[-1].is_trainable():
            train_seq[-1] = (train_seq[-1][0],
                             self._get_stop_training_wrapper(self._flow[-1],
                                                             train_seq[-1][1]))
        return train_seq

    ## Helper methods for _get_train_seq. ##

    def _get_train_function(self, nodenr):
        """Internal function factory for train.

        nodenr -- the index of the node to be trained
        """
        # This method is similar to the first part of
        # BiFlow._train_node_single_phase.
        def _train(x, msg=None):
            target = self._get_target()
            i_node = self._flow._target_to_index(target)
            ## loop until we have to go outside or complete train
            while True:
                ## execute flow before training node
                result = self._flow._execute_seq(x, msg, i_node=i_node,
                                                 stop_at_node=nodenr)
                if (isinstance(result, tuple) and len(result) == 3 and
                    result[2] is True):
                    # we have reached the training node
                    x = result[0]
                    msg = result[1]
                    i_node = nodenr  # have to update this manually
                else:
                    # flownode should be reentered later
                    return result
                ## perform node training
                if isinstance(self._flow[nodenr], BiNode):
                    result = self._flow[nodenr].train(x, msg)
                    if result is None:
                        return None  # training is done for this chunk
                else:
                    self._flow[nodenr].train(x)
                    return None
                ## training execution continues, interpret result
                if not isinstance(result, tuple):
                    x = result
                    msg = None
                    target = None
                elif len(result) == 2:
                    x, msg = result
                    target = None
                elif len(result) == 3:
                    x, msg, target = result
                else:
                    # reaching this is probably an error, leave the handling
                    # to the outer flow
                    return result
                ## check if the target is in this flow, return otherwise
                if isinstance(target, int):
                    i_node = i_node + target
                    # values of +1 and -1 beyond this flow are allowed
                    if i_node == len(self._flow):
                        if not msg:
                            return x
                        else:
                            return (x, msg)
                    elif i_node == -1:
                        return x, msg, -1
                else:
                    i_node = self._flow._target_to_index(target, i_node)
                    if not isinstance(i_node, int):
                        # target not found in this flow
                        # this is also the exit point when EXIT_TARGET is given
                        return x, msg, target
        # return the custom _train function
        return _train

    def _get_stop_training_function(self, nodenr):
        """Internal function factory for stop_training.

        nodenr -- the index of the node for which the training stops
        """
        # This method is similar to the second part of
        # BiFlow._train_node_single_phase.
        def _stop_training(msg=None):
            if isinstance(self._flow[nodenr], BiNode):
                result = self._flow[nodenr].stop_training(msg)
            else:
                # for a non-bi Node the msg is dropped
                result = self._flow[nodenr].stop_training()
            # process stop_training result
            if result is None:
                return None
            # prepare execution phase
            if not isinstance(result, tuple):
                x = result
                msg = None
                target = None
            elif len(result) == 2:
                x, msg = result
                target = None
            elif len(result) == 3:
                x, msg, target = result
            else:
                err = ("Node produced invalid return value " +
                       "for stop_training: " + str(result))
                raise BiFlowException(err)
            if isinstance(target, int):
                i_node = nodenr + target
                # values of +1 and -1 beyond this flow are allowed
                if i_node == len(self._flow):
                    return x, msg, 1
                elif i_node == -1:
                    return x, msg, -1
            else:
                i_node = self._flow._target_to_index(target, nodenr)
                if not isinstance(i_node, int):
                    # target not found in this flow
                    # this is also the exit point when EXIT_TARGET is given
                    return x, msg, target
            return self._flow._execute_seq(x, msg, i_node=i_node)
        # return the custom _stop_training function
        return _stop_training

    def _get_stop_training_wrapper(self, node, func):
        """Return wrapper for stop_training to set BiFlowNode outputdim."""
        # We have to overwrite the version from FlowNode to take care of the
        # optional return value.
        def _stop_training_wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            self.output_dim = node.output_dim
            return result
        return _stop_training_wrapper

    ### Special BiNode methods ###

    def _bi_reset(self):
        self._last_id_request = None
        for node in self._flow:
            if isinstance(node, BiNode):
                node.bi_reset()

    def _request_node_id(self, node_id):
        if self._node_id == node_id:
            return self
        for node in self._flow:
            if isinstance(node, BiNode):
                found_node = node._request_node_id(node_id)
                if found_node:
                    self._last_id_request = node_id
                    return found_node
        return None



from builtins import range
from ._tools import *

TESTYPES = [numx.dtype('d'), numx.dtype('f')]

def testCovarianceMatrix():
    mat,mix,inp = get_random_mix()
    des_cov = numx.cov(inp, rowvar=0)
    des_avg = mean(inp,axis=0)
    des_tlen = inp.shape[0]
    act_cov = utils.CovarianceMatrix()
    act_cov.update(inp)
    act_cov,act_avg,act_tlen = act_cov.fix()
    assert_array_almost_equal(act_tlen,des_tlen, decimal)
    assert_array_almost_equal(act_avg,des_avg, decimal)
    assert_array_almost_equal(act_cov,des_cov, decimal)

def testDelayCovarianceMatrix():
    dt = 5
    mat,mix,inp = get_random_mix()
    des_tlen = inp.shape[0] - dt
    des_avg = mean(inp[:des_tlen,:],axis=0)
    des_avg_dt = mean(inp[dt:,:],axis=0)
    des_cov = utils.cov2(inp[:des_tlen,:], inp[dt:,:])
    act_cov = utils.DelayCovarianceMatrix(dt)
    act_cov.update(inp)
    act_cov,act_avg,act_avg_dt,act_tlen = act_cov.fix()
    assert_array_almost_equal(act_tlen,des_tlen, decimal-1)
    assert_array_almost_equal(act_avg,des_avg, decimal-1)
    assert_array_almost_equal(act_avg_dt,des_avg_dt, decimal-1)
    assert_array_almost_equal(act_cov,des_cov, decimal-1)

def testCrossCovarianceMatrix():
    mat,mix,inp1 = get_random_mix(mat_dim=(500,5))
    mat,mix,inp2 = get_random_mix(mat_dim=(500,3))
    des_tlen = inp1.shape[0]
    des_avg1 = mean(inp1, axis=0)
    des_avg2 = mean(inp2, axis=0)
    des_cov = utils.cov2(inp1, inp2)
    act_cov = utils.CrossCovarianceMatrix()
    act_cov.update(inp1, inp2)
    act_cov, act_avg1, act_avg2, act_tlen = act_cov.fix()
    assert_almost_equal(act_tlen,des_tlen, decimal-1)
    assert_array_almost_equal(act_avg1,des_avg1, decimal-1)
    assert_array_almost_equal(act_avg2,des_avg2, decimal-1)
    assert_array_almost_equal(act_cov,des_cov, decimal-1)

def testdtypeCovarianceMatrix():
    for type in TESTYPES:
        mat,mix,inp = get_random_mix(type='d')
        cov = utils.CovarianceMatrix(dtype=type)
        cov.update(inp)
        cov,avg,tlen = cov.fix()
        assert_type_equal(cov.dtype,type)
        assert_type_equal(avg.dtype,type)

def testdtypeDelayCovarianceMatrix():
    for type in TESTYPES:
        dt = 5
        mat,mix,inp = get_random_mix(type='d')
        cov = utils.DelayCovarianceMatrix(dt=dt, dtype=type)
        cov.update(inp)
        cov,avg,avg_dt,tlen = cov.fix()
        assert_type_equal(cov.dtype,type)
        assert_type_equal(avg.dtype,type)
        assert_type_equal(avg_dt.dtype,type)

def testdtypeCrossCovarianceMatrix():
    for type in TESTYPES:
        mat,mix,inp = get_random_mix(type='d')
        cov = utils.CrossCovarianceMatrix(dtype=type)
        cov.update(inp, inp)
        cov,avg1,avg2,tlen = cov.fix()
        assert_type_equal(cov.dtype,type)
        assert_type_equal(avg1.dtype,type)
        assert_type_equal(avg2.dtype,type)

def testRoundOffWarningCovMatrix():
    import warnings
    warnings.filterwarnings("error",'.*',mdp.MDPWarning)
    for type in ['f','d']:
        inp = uniform((1,2))
        cov = utils.CovarianceMatrix(dtype=type)
        cov._tlen = int(1e+15)
        cov.update(inp)
        try:
            cov.fix()
            assert False, 'RoundOff warning did not work'
        except mdp.MDPWarning:
            pass
    # hope to reset the previous state...
    warnings.filterwarnings("once",'.*',mdp.MDPWarning)

def testMultipleCovarianceMatricesDtypeAndFuncs():
    for type in TESTYPES:
        dec = testdecimals[type]
        res_type = _MultipleCovarianceMatrices_funcs(type,dec)
        assert_type_equal(type,res_type)


def _MultipleCovarianceMatrices_funcs(dtype, decimals):
    def assert_all(des,act, dec=decimals):
        # check list of matrices equals multcov array
        for x in range(nmat):
            assert_array_almost_equal_diff(des[x],act.covs[:,:,x],dec)

    def rotate(mat,angle,indices):
        # perform a givens rotation of a single matrix
        [i,j] = indices
        c, s = numx.cos(angle), numx.sin(angle)
        mat_i, mat_j = mat[:,i].copy(), mat[:,j].copy()
        mat[:,i], mat[:,j] = c*mat_i-s*mat_j, s*mat_i+c*mat_j
        mat_i, mat_j = mat[i,:].copy(), mat[j,:].copy()
        mat[i,:], mat[j,:] = c*mat_i-s*mat_j, s*mat_i+c*mat_j
        return mat.copy()

    def permute(mat,indices):
        # permute rows and cols of a single matrix
        [i,j] = indices
        mat_i, mat_j = mat[:,i].copy(), mat[:,j].copy()
        mat[:,i], mat[:,j] = mat_j, mat_i
        mat_i, mat_j = mat[i,:].copy(), mat[j,:].copy()
        mat[i,:], mat[j,:] = mat_j, mat_i
        return mat.copy()

    dim = 7
    nmat = 13
    # create mult cov mat
    covs = [uniform((dim,dim)).astype(dtype) for x in range(nmat)]
    mult_cov = mdp.utils.MultipleCovarianceMatrices(covs)
    assert_equal(nmat,mult_cov.ncovs)
    # test symmetrize
    sym_covs = [0.5*(x+x.T) for x in covs]
    mult_cov.symmetrize()
    assert_all(sym_covs,mult_cov)
    # test weight
    weights = uniform(nmat)
    w_covs = [weights[x]*sym_covs[x] for x in range(nmat)]
    mult_cov.weight(weights)
    assert_all(w_covs,mult_cov)
    # test rotate
    angle = uniform()*2*numx.pi
    idx = numx_rand.permutation(dim)[:2]
    rot_covs = [rotate(x,angle,idx) for x in w_covs]
    mult_cov.rotate(angle,idx)
    assert_all(w_covs,mult_cov)
    # test permute
    per_covs = [permute(x,idx) for x in rot_covs]
    mult_cov.permute(idx)
    assert_all(per_covs,mult_cov)
    # test transform
    trans = uniform((dim,dim))
    trans_covs = [mult(mult(trans.T,x),trans) for x in per_covs]
    mult_cov.transform(trans)
    assert_all(trans_covs,mult_cov)
    # test copy
    cp_mult_cov = mult_cov.copy()
    assert_array_equal(mult_cov.covs,cp_mult_cov.covs)
    # check that we didn't got a reference
    mult_cov[0][0,0] = 1000
    assert int(cp_mult_cov[0][0,0]) != 1000
    # return dtype
    return mult_cov.covs.dtype

def testMultipleCovarianceMatricesTransformations():
    def get_mult_covs(inp,nmat):
        # return delayed covariance matrices
        covs = []
        for delay in range(nmat):
            tmp = mdp.utils.DelayCovarianceMatrix(delay)
            tmp.update(inp)
            cov,avg,avg_dt,tlen = tmp.fix()
            covs.append(cov)
        return mdp.utils.MultipleCovarianceMatrices(covs)
    dim = 7
    nmat = 13
    angle = uniform()*2*numx.pi
    idx = numx_rand.permutation(dim)[:2]
    inp = uniform((100*dim,dim))
    rot_inp, per_inp = inp.copy(), inp.copy()
    # test if rotating or permuting the cov matrix is equivalent
    # to rotate or permute the sources.
    mdp.utils.rotate(rot_inp,angle,idx)
    mdp.utils.permute(per_inp,idx,rows=0,cols=1)
    mcov = get_mult_covs(inp, nmat)
    mcov2 = mcov.copy()
    mcov_rot = get_mult_covs(rot_inp, nmat)
    mcov_per = get_mult_covs(per_inp, nmat)
    mcov.rotate(angle,idx)
    mcov2.permute(idx)
    assert_array_almost_equal_diff(mcov.covs, mcov_rot.covs, decimal)
    assert_array_almost_equal_diff(mcov2.covs, mcov_per.covs, decimal)



# global hooks for py.test
import tempfile
import os
import shutil
import glob
import mdp
import py.test

_err_str = """
IMPORTANT: some tests use random numbers. This could
occasionally lead to failures due to numerical degeneracies.
To rule this out, please run the tests more than once.
If you get reproducible failures please report a bug!
"""

def pytest_configure(config):
    seed = config.getvalue("seed")
    # if seed was not set by the user, we set one now
    if seed is None or seed == ('NO', 'DEFAULT'):
        config.option.seed = int(mdp.numx_rand.randint(2**31-1))

def pytest_unconfigure(config):
    # remove garbage created during tests
    # note that usage of TemporaryDirectory is not enough to assure
    # that all garbage is removed, expacially because we use subprocesses
    shutil.rmtree(py.test.mdp_tempdirname, ignore_errors=True)
    # if pp was monkey-patched, remove any stale pp4mdp directories
    if hasattr(mdp.config, 'pp_monkeypatch_dirname'):
        monkey_dirs = os.path.join(mdp.config.pp_monkeypatch_dirname,
                                   mdp.parallel.pp_support.TEMPDIR_PREFIX) 
        [shutil.rmtree(d, ignore_errors=True) for d in glob.glob(monkey_dirs+'*')]

def pytest_runtest_setup(item):
    # set random seed before running each test
    # so that a failure in a test can be reproduced just running
    # that particular test. if this was not done, you would need
    # to run the whole test suite again
    mdp.numx_rand.seed(item.config.option.seed)

def pytest_addoption(parser):
    """Add random seed option to py.test.
    """
    parser.addoption('--seed', dest='seed', type=int, action='store',
                     help='set random seed')

def pytest_report_header(config):
    # report the random seed before and after running the tests
    return '%s\nRandom Seed: %d\n' % (mdp.config.info(), config.option.seed)

def pytest_terminal_summary(terminalreporter):
    # add a note about error due to randomness only if an error or a failure
    # occured
    t = terminalreporter
    t.write_sep("=", "NOTE")
    t.write_line("%s\nRandom Seed: %d" % (mdp.config.info(),
                                          t.config.option.seed))
    if 'failed' in t.stats or 'error' in t.stats:
        t.write_line(_err_str)

def pytest_namespace():
    # get temporary directory to put temporary files
    # will be deleted at the end of the test run
    dirname = tempfile.mkdtemp(suffix='.tmp', prefix='MDPtestdir_')
    return dict(mdp_tempdirname=dirname)




import mdp
import inspect
import py.test
X = mdp.numx_rand.random(size=(500,5))

def get_signature(func):
    regargs, varargs, varkwargs, defaults = inspect.getargspec(func)
    return inspect.formatargspec(regargs, varargs, varkwargs, defaults,
                                 formatvalue=lambda value: "")[1:-1]

def teardown_function(function):
    """Deactivate all extensions and remove testing extensions."""
    mdp.deactivate_extensions(mdp.get_active_extensions())
    for key in mdp.get_extensions().copy():
        if key.startswith("__test"):
            del mdp.get_extensions()[key]

def test_signatures_same_no_arguments():
    class AncestorNode(mdp.Node):
        def _train(self, x, foo2=None):
            self.foo2 = None
    class ChildNode(AncestorNode):
        def _train(self, x, foo=None):
            self.foo = foo
    cnode = ChildNode()
    assert get_signature(cnode.train) == 'self, x, foo'
    assert get_signature(cnode._train) == 'self, x, foo'
    cnode.train(X, foo=42)
    assert cnode.foo == 42
    py.test.raises(AttributeError, 'cnode.foo2')

def test_signatures_more_arguments():
    class AncestorNode(mdp.Node):
        def _train(self, x):
            self.foo2 = None
    class ChildNode(AncestorNode):
        def _train(self, x, foo=None):
            self.foo = foo
    cnode = ChildNode()
    assert get_signature(cnode.train) == 'self, x, foo'
    assert get_signature(cnode.train._undecorated_) == 'self, x, *args, **kwargs'
    assert get_signature(cnode._train) == 'self, x, foo'
    # next two lines should give the same:
    cnode.train._undecorated_(cnode, X, foo=42)
    cnode.train(X, foo=42)
    assert cnode.foo == 42
    py.test.raises(AttributeError, 'cnode.foo2')

def test_signatures_less_arguments():

    class AncestorNode(mdp.Node):
        def _train(self, x, foo=None):
            self.foo = None

    class ChildNode(AncestorNode):
        def _train(self, x):
            self.moo = 3

    cnode = ChildNode()
    assert get_signature(cnode.train) == 'self, x'
    assert get_signature(cnode.train._undecorated_) == 'self, x, *args, **kwargs'
    assert get_signature(cnode._train) == 'self, x'

    # next two lines should give the same:
    cnode.train._undecorated_(cnode, X)
    cnode.train(X)
    assert cnode.moo == 3
    py.test.raises(AttributeError, 'cnode.foo')

def test_simple_extension():

    class TestExtensionNode(mdp.ExtensionNode, mdp.nodes.IdentityNode):
        extension_name = "__test"
        def execute(self, x):
            self.foo = 42
            return self._non_extension_execute(x)

    class Dummy(mdp.nodes.IdentityNode):
        def _execute(self, x):
            return 42

    node = mdp.nodes.IdentityNode()
    assert mdp.numx.all(node.execute(X) == X)
    assert not hasattr(node,'foo')

    with mdp.extension("__test"):
        assert mdp.numx.all(node.execute(X) == X)
        assert hasattr(node,'foo')

    node = Dummy()
    assert not hasattr(node,'foo')
    assert node.execute(X) == 42

    with mdp.extension("__test"):
        assert node.execute(X) == 42
        assert hasattr(node,'foo')



"""
Script to be called to start a network server.

It differs from calling ppserver.py mainly in that it allows to add paths to
sys.path. So it acts like a wrapper for the server initialization.
The paths are passed via sys.argv[5:]. The first four arguments are
port, timeout, secret, n_workers.
"""
from __future__ import print_function

import sys

def main():
    port, timeout, secret, n_workers = sys.argv[1:5]
    port = int(port)
    timeout = int(timeout)
    n_workers = int(n_workers)
    sys_paths = sys.argv[5:]
    for sys_path in sys_paths:
        sys.path.append(sys_path)
    import ppserver
    ## initialization code as in ppserver.py
    server = ppserver._NetworkServer(ncpus=n_workers,
                                     port=port,
                                     secret=secret,
                                     timeout=timeout)
    print("Server is ready.")
    sys.stdout.flush()
    server.listen()

if __name__ == "__main__":
    main()
